#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

export YELLOW='\033[0;33m'
export RED='\033[0;31m'
export NC='\033[0m' # No Color

function error() {
  echo -e "${RED}[ERROR][$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*${NC}\n" >&2
}

function info() {
  echo "[INFO] [$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*" >&2
}

function warn() {
  echo "${YELLOW}[WARN] [$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*${NC}" >&2
}

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi

function notify_pushgateway() {
  local metric_name="${1}"
  local result="${2}"
  local volume="${3-none}"
  local pushgateway_url="http://pushgateway-prometheus-pushgateway.cattle-monitoring-system.svc.cluster.local"
  local pushgateway_port="9091"
  url="${pushgateway_url}:${pushgateway_port}/metrics/job/backup-svc"
  if [[ "${volume}" != "none" ]]; then
    url="${url}/volume/${volume}"
  fi
  info "Notify alert for ${metric_name} : ${result} in pushgateway."
  http_response=$(echo "${metric_name} ${result}" | curl --data-binary @-  -w "%{http_code}" "${url}") || true

  if [[ "${http_response}" != "200" ]]; then
    error "Failed to send alerts to Pushgateway, StatusCode: ${http_response}"
  else
    info "Successfully notified alert for ${metric_name} in pushgateway."
  fi
}

function argocd_login() {
  local try=0
  local maxtry=10
  local argocd_server argocd_server_name argocd_server_port argocd_namespace argocd_endpoint

  while (( try != maxtry )) ; do
    try=$((try+1))
    (( try == 1 )) || sleep 20
    info "Trying to log in to argo...${try}/${maxtry}";

    argocd_server=$(kubectl get svc -A -l app.kubernetes.io/name=argocd-server -o json)
    argocd_server_name="$(echo "${argocd_server}" | jq -r '.items[0].metadata.name')"
    argocd_server_port="$(echo "${argocd_server}"  | jq -r '.items[0].spec.ports[] | select (.name=="http") | .port')"
    argocd_namespace="$(echo "${argocd_server}" | jq -r '.items[0].metadata.namespace')"
    argocd_password=$(kubectl -n "${argocd_namespace}" get secret argocd-admin-password -o jsonpath="{.data.password}" | base64 -d)
    argocd_endpoint="${argocd_server_name}.${argocd_namespace}.svc.cluster.local:${argocd_server_port}"

    info "ArgoCD SERVER: $argocd_endpoint"

    argocd login \
        "${argocd_endpoint}" \
        --username admin --password "${argocd_password}" \
        --plaintext \
        && info "Login argocd successfully." && ARGOCD_HEALTH="healthy" && return 0
  done
  error "Failed to login argocd"
  ARGOCD_HEALTH="unhealthy"
}

function check_unhealthy_apps() {
  local service_name argocd_binary applications_name all_sorted_application
  argocd_binary="$1"
  service_name="$2"

  info "Start validating ${service_name} child's application health."
  applications_name=$(${argocd_binary} app resources "${service_name}" | grep Application | awk '{print $4}')
  if [[ "${applications_name}" == "" ]]; then
    error "Failed to fetch list of app resources for ${service_name}."
    APPLICATION_HEALTH="unhealthy"
    return 0
  fi
  all_sorted_application=[]
  for application_name in $applications_name;
  do
      app=$(${argocd_binary} app get "${application_name}" -o json | jq ".|{name: .metadata.name, wave: .metadata.annotations.\"argocd.argoproj.io/sync-wave\", syncStatus: .status.sync.status, healthStatus: .status.health.status, phaseStatus: .status.operationState.phase}|select((.healthStatus!=\"Healthy\") or (.phaseStatus!=\"Succeeded\"))")
      all_sorted_application=$(jq ". += [${app}]" <<< "${all_sorted_application}")
  done

  all_sorted_application=$(echo "${all_sorted_application}" | jq -s -c ".[] | sort_by(.wave)")

  if [[ "${all_sorted_application}" == "[]" ]]; then
    info "All application under ${service_name} are healthy"
    APPLICATION_HEALTH="healthy"
    return 0
  fi
  error "${all_sorted_application} are found in unhealthy state."
  APPLICATION_HEALTH="unhealthy"
}

function validate_longhorn_pods_health() {
  local crashloopbackoffcount
  info "Validating longhorn pods health.."
  crashloopbackoffcount=$(kubectl get pods -n longhorn-system | grep -c "CrashLoopBackOff") || true
  if [[ "${crashloopbackoffcount}" == 0 ]]; then
    info "Validated longhorn pods are healthy"
    LONGHORN_HEALTH="healthy"
    return 0
  fi
  error "Longhorn pods are not healthy"
  LONGHORN_HEALTH="unhealthy"
}

function validate_argocd_health() {
  local argocd_svc
  local argocd_ip
  local argocd_port

  # get argocd server ip & port
  argocd_svc=$(kubectl -n argocd get svc argocd-server -o=json | jq -c '.spec| {ip: .clusterIP, port: .ports[0].port}')
  argocd_ip=$(echo "${argocd_svc}"| jq -r ".ip")
  argocd_port=$(echo "${argocd_svc}"| jq -r ".port")

  info "Validating argocd health.."
  local argocd_url
  local argocd_result=""
  argocd_url="http://${argocd_ip}:${argocd_port}"
  argocd_result=$(curl "${argocd_url}/api/version" --connect-timeout 60) || true
  if [[ -z "${argocd_result}" || "${argocd_result}" != *"Version"* ]]; then
    error "Validated argocd not in healthy state"
    ARGOCD_HEALTH="unhealthy"
    return 0
  fi
  info "Argocd is healthy."
  ARGOCD_HEALTH="healthy"
}

function check_cluster_health() {
  LONGHORN_HEALTH="unhealthy"
  ARGOCD_HEALTH="unhealthy"
  # check health of longhorn and argocd pods
  validate_longhorn_pods_health
  validate_argocd_health

  if [[ "${LONGHORN_HEALTH}" != "healthy" || "${ARGOCD_HEALTH}" != "healthy" ]];then
    error "Cluster is unhealthy, longhorn: ${LONGHORN_HEALTH}, argocd: ${ARGOCD_HEALTH}"
    CLUSTER_HEALTH="unhealthy"
    return 0
  fi
  # login to argocd
  argocd_login || error "Failed to login to argocd"
  if [[ "${ARGOCD_HEALTH}" != "healthy" ]]; then
    CLUSTER_HEALTH="unhealthy"
    return 0
  fi

  APPLICATION_HEALTH="unhealthy"
  # check argocd health of all apps
  check_unhealthy_apps "argocd" "fabric-installer"
  fabric_apps_health_status=$APPLICATION_HEALTH
  if [[ "${fabric_apps_health_status}" != "healthy" ]]; then
    CLUSTER_HEALTH="unhealthy"
    error "Cluster is unhealthy, fabric component: ${fabric_apps_health_status}"
    return 0
  fi

  # APPLICATION_HEALTH="unhealthy"
  # check_unhealthy_apps "argocd" "uipath"
  # service_apps_health_status=$APPLICATION_HEALTH
  # if [[ "${service_apps_health_status}" != "healthy" ]]; then
  #   CLUSTER_HEALTH="unhealthy"
  #   error "Cluster is unhealthy, service component: ${service_apps_health_status}"
  #   return 0
  # fi
  
  CLUSTER_HEALTH="healthy"
  info "Cluster is healthy."
}

function update_uipath_backup(){
  local bool=$1
  bool_base64enc=$(echo "${bool}"|base64)
  kubectl -n uipath-infra patch secret uipath-backup --type='json' -p="[{\"op\" : \"replace\" ,\"path\" : \"/data/backup_enabled\" ,\"value\" : \"${bool_base64enc}\"}]"
}

function check_backup_enabled() {
  # if flag is not found or true then backup is enabled else disabled
  local flag="true"
  if kubectl -n uipath-infra get secret uipath-backup ; then
    flag=$(kubectl -n uipath-infra get secret uipath-backup -o json| jq -r ".data.backup_enabled"|base64 -d)
  fi
  export BACKUP_ENABLED="${flag}"
}

function create_recurring_job() {
  local cron_job_expression
  cron_job_expression="${1}"

  cat <<EOF | kubectl apply -f -
apiVersion: longhorn.io/v1beta1
kind: RecurringJob
metadata:
  name: uipath-backup
  namespace: longhorn-system
spec:
  concurrency: 4
  cron: "${cron_job_expression}"
  groups:
  - uipath-backup
  retain: 1
  task: backup
EOF

  # validate pvc cronjob for backup created
  local try=0
  local maxtry=10
  local exists=""
  while (( try < maxtry ));
  do
    exists=$(kubectl -n longhorn-system get cronjobs uipath-backup) || true
    if [[ -n "${exists}" ]]; then
      info "recurring cronjob uipath-backup created successfully"
      break;
    fi

    warn "recurring cronjob uipath-backup not created yet, ${try}/${maxtry}..."; sleep 5
    try=$((try + 1))
  done

  if [[ -z "${exists}" ]];then
    error "failed to create recurring job for uipath-backup"
    return 0
  fi
}

function delete_recurring_job() {
  backup_job_name="${1}"

  # validate pvc cronjob for backup created
  local try=0
  local maxtry=10
  local exists=""
  while (( try < maxtry ));
  do
    # delete recurring job
    kubectl -n longhorn-system delete recurringjob "${backup_job_name}" --force || true
    kubectl -n longhorn-system delete cronjobs "${backup_job_name}" --force || true

    exists=$(kubectl -n longhorn-system get cronjobs "${backup_job_name}") || true
    if [[ -z "${exists}" ]]; then
      info "recurring cronjob ${backup_job_name} deleted successfully"
      break;
    fi

    warn "recurring cronjob ${backup_job_name} not deleted yet, ${try}/${maxtry}..."; sleep 5
    try=$((try + 1))
    exists=""
  done

  if [[ -n "${exists}" ]];then
    error "failed to delete recurring job for ${backup_job_name}"
    return 0
  fi
}


function enable_backup() {
  local backup_interval cronexpression
  #shellcheck disable=SC2153
  backup_interval="${BACKUP_INTERVAL, 15}"
  cronexpression="*/${backup_interval} * * * *"
  # create recurring job
  create_recurring_job "${cronexpression}"
  # remove the flag file
  update_uipath_backup "true"
}

function disable_backup() {
  # delete recurring job
  delete_recurring_job "uipath-backup"
  # add the flag file
  update_uipath_backup "false"
}

function check_nfs_server_connection() {
  local endpoint="${NFS_SERVER_FQDN}"
  local port="${NFS_SERVER_PORT}"

  local connection
  
  # shellcheck disable=SC2086
  # validate connection
  connection=$(nc -vz "${endpoint}" "${port}" 2>&1)
  if [[ ! (${connection} == *"Connected"* || ${connection} == *"succeeded"*) ]]; then
    NFS_MOUNT_STATUS="unhealthy"
    error "Failed to connect to NFS Server at ${endpoint}"
    notify_pushgateway "nfs_server_status" "0"
    return 0
  fi

  NFS_MOUNT_STATUS="healthy"
  notify_pushgateway "nfs_server_status" "1"
  info "Validated NFS Server connection successfully."
}

function check_volume_backup_status() {
  local volume backup_job_logs job_created job_completed
  volume=${1}
  backup_job_logs=${2}
  job_created=$(echo "${backup_job_logs}"|grep "${volume}"|grep "Creating job")
  if [[ -z "${job_created}" ]]; then
    notify_pushgateway "volume_backup_status" "0" "${volume//-/_}"
    return 0
  fi
  job_completed=$(echo "${backup_job_logs}"|grep "${volume}"|grep "Finish running recurring backup")
  if [[ -z "${job_completed}" ]]; then
    notify_pushgateway "volume_backup_status" "0" "${volume//-/_}"
    return 0
  fi
  notify_pushgateway "volume_backup_status" "1" "${volume//-/_}"
}

function check_schedule_backup_status() {
  local pod_exist pod_name backup_job_logs
  pod_exist=$(kubectl get pods -n longhorn-system -l "recurring-job.longhorn.io=uipath-backup" --sort-by=.metadata.creationTimestamp --field-selector=status.phase==Succeeded)
  if [[ -z "${pod_exist}" ]]; then
    return 0
  fi
  pod_name=$(kubectl get pods -n longhorn-system -l "recurring-job.longhorn.io=uipath-backup" --sort-by=.metadata.creationTimestamp --field-selector=status.phase==Succeeded -o 'jsonpath={.items[-1].metadata.name}')
  if [[ -z "${pod_name}" ]]; then
    return 0
  fi

  backup_job_logs=$(kubectl -n longhorn-system logs "${pod_name}" ) || true
  if [[ -z "${backup_job_logs}" ]]; then
    notify_pushgateway "volume_backup_status" "0"
    return 0
  fi

  notify_pushgateway "volume_backup_status" "1"
  # check latest volume backup
  volumes_list=$(kubectl -n longhorn-system get volumes.longhorn.io -l "recurring-job-group.longhorn.io/uipath-backup=enabled" -o json | jq -r ".items[].metadata.name")
  for volume in ${volumes_list}; do
    check_volume_backup_status "${volume}" "${backup_job_logs}"
  done
}

function main() {
  while true; do
    # check schedule backup status
    check_schedule_backup_status

    NFS_MOUNT_STATUS="unhealthy"
    # check_nfs_server_alert
    check_nfs_server_connection

    CLUSTER_HEALTH="unhealthy"
    # check cluster health
    check_cluster_health

    if [[ "${NFS_MOUNT_STATUS}" == "healthy" && "${CLUSTER_HEALTH}" == "healthy" ]]; then
      check_backup_enabled
      if [[ "${BACKUP_ENABLED}" != "true" ]]; then
        enable_backup
      else
        info "Backup is already enabled."
      fi
      # notify disable backup
      notify_pushgateway "backup_status" "1"
    else
      check_backup_enabled
      if [[ "${BACKUP_ENABLED}" == "true" ]]; then
        disable_backup
      else
        info "Backup is already disabled."
      fi
      # notify disable backup
      notify_pushgateway "backup_status" "0"
    fi
    unset BACKUP_ENABLED
    # validate cluster every 300 sec
    sleep 300
  done
}

main