#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

########
# Utils module for different script functions
########

# Redirect /dev/null to INSTALLER_LOGFILE if set
dev_null="/dev/null"
[[ -n "${INSTALLER_LOGFILE:-}" ]] && dev_null="${INSTALLER_LOGFILE}"

# TODO: this needs to be based on target env type so we don't have to patch the release branch to repoint to customer instance
TELEMETRY_INSTRUMENTATION_KEY=5da85bdb-6422-49b5-a6fa-8d3a0907a0f0
TELEMETRY_FQDN=dc.services.visualstudio.com
TELEMETRY_URL=https://${TELEMETRY_FQDN}/v2/track
function send_telemetry_unsafe() {
  curl -sf "${TELEMETRY_URL}" -k -X POST \
    -H 'Content-Type: application/json' -m 30 -d "${1}" >> "${dev_null}" 2>&1
}

function send_telemetry() {
  send_telemetry_unsafe "$@" || true
}

function ensure_machine_hash() {
  if [[ "${MACHINE_HASH:-none}" == "none" ]]; then
    MACHINE_HASH="$(uname -a | sha256sum -t - | cut -d' ' -f1)"
  fi

  export MACHINE_HASH
}

function service_exists() {
    local service_name=$1
    if [[ $(systemctl list-units --all -t service --full --no-legend "$service_name.service" | cut -f1 -d' ') == $service_name.service ]]; then
      echo "true"
    else
      echo "false"
    fi
}

function read_cluster_id() {
  local app_details
  local try=0
  local maxtry=3
  local status="notready"
  while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
          try=$((try+1))
          app_details=$(argocd app get uipath --show-params 2>/dev/null) && status="ready"
          info "waiting for the uipath argo app details ...${try}/${maxtry}" &&  sleep 5;
  done

  [[ ${status} == "ready" ]] && CLUSTER_ID=$(echo "${app_details}" | grep telemetry.clusterId | awk '{print $2}')
  [[ ${status} == "notready" ]] && CLUSTER_ID="none"

  if [[ -z ${CLUSTER_ID} ]]; then
    error "Failed to get the telemetry ClusterId"
  else
    info "[Telemetry] Got the telemetry clusterId: ${CLUSTER_ID}"
  fi
}

# For k8s operation, node name should be in lowercase.
# To avoid usage of `hostname` command for k8s operation, use fetch_hostname_lowercase to get node name
function fetch_hostname_lowercase() {
  local node_name
  node_name=$(xargs -0 < "/proc/$(pgrep kubelet)/cmdline" | tr ' ' '\n' | grep '\-\-hostname\-override' | cut -d'=' -f2)
  [[ -z "${node_name}" ]] && node_name=$(fetch_hostname | awk '{print tolower($0)}')
  echo "${node_name}"
}

# fetch_hostname return hostname as it is.
# For k8s operation, use fetch_hostname_lowercase to get node name
function fetch_hostname() {
  hostname
}

function send_checkpoint_telemetry() {
  local telemetry_content

  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    return;
  fi

  local now
  local checkpoint

  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)
  checkpoint=${1}

  ensure_machine_hash

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Checkpoint",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer",
                "InstallId": "${INSTALL_ID:-none}",
                "MachineHash": "${MACHINE_HASH}",
                "ClusterToken" : "${RKE_TOKEN:-none}",
                "ClusterId": "${CLUSTER_ID:-none}",
                "Checkpoint": "${checkpoint}"
            }
        }
    }
}
EOF
  }

  telemetry_content="$(data)"
  send_telemetry "${telemetry_content}"
}

function send_error_telemetry() {
  local telemetry_content

  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    return;
  fi

  local now
  local origin
  local message

  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)
  origin=${1}
  message=${*:2}

  ensure_machine_hash

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Error",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer",
                "InstallId": "${INSTALL_ID:-none}",
                "MachineHash": "${MACHINE_HASH}",
                "ClusterToken" : "${RKE_TOKEN:-none}",
                "ClusterId": "${CLUSTER_ID:-none}",
                "Origin": "${origin}",
                "Message": "${message}"
            }
        }
    }
}
EOF
  }

  telemetry_content="$(data)"
  send_telemetry "${telemetry_content}"
}

function get_message_with_errorcode() {
  local err_code=$1
  local type=$2
  err_message=$(jq -r ".${type}|select( has (\"${err_code}\")).\"${err_code}\"" "${MODULES_DIR}"/errcode_message_mapping.json)
  # update positional args
  local i=0
  #shellcheck disable=SC2048
  for arg in ${*:3}; do
    err_message=${err_message//\{$i\}/${arg}}
    i=$((i+1))
  done

  echo "${err_message}"
}

function error_with_code() {
  local err_code=$1
  local error_message
  error_message="$(get_message_with_errorcode "${err_code}" "error" "${*:2}")"

  echo -e "\e[0;31m[ERROR][$(date +'%Y-%m-%dT%H:%M:%S%z')]:\e[0m ${err_code} : ${error_message}" >&2

  [[ -n "${INSTALLER_LOGFILE:-}" ]] && echo -e "\e[0;33mMore details can be found in ${INSTALLER_LOGFILE}\e[0m"

  send_error_telemetry "$0" "${error_message}"
  exit 1
}

function warn_with_code() {
  local warn_code=$1
  local warn_message
  warn_message="$(get_message_with_errorcode "${warn_code}" "warn" "${*:2}")"

  echo -e "\e[0;33m[WARN][$(date +'%Y-%m-%dT%H:%M:%S%z')]:\e[0m ${warn_code} : ${warn_message}" >&2
}

function error() {
  echo -e "\e[0;31m[ERROR][$(date +'%Y-%m-%dT%H:%M:%S%z')]:\e[0m $*" >&2
  [[ -n "${INSTALLER_LOGFILE:-}" ]] && echo -e "\e[0;33mMore details can be found in ${INSTALLER_LOGFILE}\e[0m"
  send_error_telemetry "$0" "$1"
  exit 1
}

function error_without_exit() {
  echo -e "\e[0;31m[ERROR][$(date +'%Y-%m-%dT%H:%M:%S%z')]:\e[0m $*" >&2
}

function info() {
  echo "[INFO] [$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*"
}

function warn() {
  echo -e "\e[0;33m[WARN] [$(date +'%Y-%m-%dT%H:%M:%S%z')]:\e[0m $*" >&2
}

function display_usage() {
  echo "usage: $(basename "$0") [-hd] -i infile -o outfile"
  echo "  -d          Run in debug"
  echo "  -h          Display help"
  echo "  -i infile   Specify input json file"
  echo "  -o outfile  Specify output json file"
}

########
# Parse command line arguments. Exit if illegal arguments are passed
# Arguments:
#   1: String to check
#   2: Error message
# Outputs:
#   INFILE and OUTFILE are initialized with their respective values
########
function exit_if_string_empty() {
  if [[ -z "$1" ]]; then
    error "$2"
  fi
}

########
# Parse command line arguments. Exit if illegal arguments are passed
# Globals:
#   INFILE
#   OUTFILE
# Outputs:
#   INFILE and OUTFILE are initialized with their respective values
########
function parse_args() {
  INFILE=""
  OUTFILE=""

  OFFLINE_INSTALLER=""
  while getopts 'hdi:o:f:t:' flag "$@"; do
    case "${flag}" in
    d)
      info "Running in debug"
      set -x
      ;;
    i)
      info "Using input parameters file: ${OPTARG}"
      INFILE=${OPTARG}
      ;;
    f)
      info "Using defaults parameters file: ${OPTARG}"
      DEFAULTSFILE=${OPTARG}
      ;;
    t)
      info "Using offlineInstaller parmeter: ${OPTARG}"
      # shellcheck disable=SC2034
      OFFLINE_INSTALLER=${OPTARG}
      ;;
    h)
      display_usage
      exit 0
      ;;
    o)
      info "Using output parameters file: ${OPTARG}"
      OUTFILE=${OPTARG}
      ;;
    :)
      error "Invalid option: ${OPTARG} requires an argument."
      ;;
    *)
      error "Unexpected option ${flag}"
      ;;
    esac
  done
  if [[ -z "${INFILE}" || -z "${OUTFILE}" ]]; then
    display_usage
    error "Mandatory parameters not found"
  fi
  if [[ ! -f "${INFILE}" ]]; then
    error "File ${INFILE} not found"
  fi
  shift $((OPTIND - 1))
}

########
# Merge defaults.json and ${INFILE} into merged_input.json
# Globals:
#   DEFAULTSFILE
# Outputs:
#   Generates "merged_input.json"
########
function merge_default_values() {
  if [[ ! -f "${DEFAULTSFILE}" ]]; then
    error "File ${DEFAULTSFILE} not found"
    exit 1
  fi
  if [ -n "${INFILE}" ];
  then
    jq '.' "${INFILE}" >/dev/null 2>> "${dev_null}" || error "Input file ${INFILE} is not a valid json  and cannot be parsed"
    jq '.' "${DEFAULTSFILE}" >/dev/null 2>> "${dev_null}" || error "Defaults file ${DEFAULTSFILE} is not a valid json and cannot be parsed"
    jq -s '.[0] * .[1]' "${DEFAULTSFILE}" "${INFILE}" > "${MERGEDINPUTFILE}"
  else
    cat "${DEFAULTSFILE}" > "${MERGEDINPUTFILE}"
  fi
  INFILE="${MERGEDINPUTFILE}"
}

#######
# Override HA flag to false if 3 master nodes are not available
#######
function check_available_masters() {
    AVAILABLE_MASTER_NODE_COUNT=$(kubectl get nodes |grep master|grep -wc "Ready")
    # shellcheck disable=SC2153
    if [ "${PROFILE}" == "ha" ] && [ "${AVAILABLE_MASTER_NODE_COUNT}" -lt 3 ]; then
      error_with_code "13005"
    fi
}

function get_fabric_install_paths() {
  local BASE_DIR=$1
  readonly BINARY_DIR="${BASE_DIR}/bin"
  export BINARY_DIR
  readonly VENDOR_DIR="${BASE_DIR}/vendor"
  export VENDOR_DIR
}

function detect_os() {
  # shellcheck disable=SC1091
  OS_ID=$(source /etc/os-release ; echo "$ID")
  # shellcheck disable=SC1091
  OS_VERSION=$(source /etc/os-release ; echo "$VERSION_ID")
  OS_MAJOR_VERSION=$(echo "$OS_VERSION" | cut -d. -f1)
  OS_MINOR_VERSION=$(echo "$OS_VERSION" | cut -d. -f2)
  export OS_ID
  export OS_VERSION
  export OS_MAJOR_VERSION
  export OS_MINOR_VERSION
}

########
# Load OS-specific utils.sh
########
function load_os_utils() {
  detect_os
  local modules_directory
  modules_directory=$1
  if [ "$OS_ID" == "ubuntu" ] || [ "$OS_ID" == "debian" ]; then
    #shellcheck disable=SC1090,SC1091
    source "${modules_directory}/debian-utils.sh"
  elif [ "$OS_ID" == "rhel" ]; then
    #shellcheck disable=SC1090,SC1091
    source "${modules_directory}/rhel-utils.sh"
  else
    warn "OS not supported. Skipping OS-specific module loading."
  fi
  info "Detected OS: $OS_ID MajorVersion: $OS_MAJOR_VERSION MinorVersion: $OS_MINOR_VERSION"
}

########
# Check if package is installed or command is present
# Uses check_package_installed from os-specific modules loaded above
########
function check_package_or_command() {
  local COMMAND_EXISTS=1
  local PACKAGE_EXISTS=1
  local PATHS
  local IFS=':'
  PATHS="$(whereis "$1")"
  # shellcheck disable=SC2162
  read -a PATHARR <<<"$PATHS"
  if [ ${#PATHARR[@]} -gt "1" ]; then
    COMMAND_EXISTS=0
  fi

  check_package_installed "$1" && PACKAGE_EXISTS=0 || PACKAGE_EXISTS=1

  if [ $PACKAGE_EXISTS == 0 ] || [ $COMMAND_EXISTS == 0 ]; then
    return 0
  else
    return 1
  fi
}

########
# Check whether a list of packages and/or commands are available
# Usage:
# required_pkgs=(jq curl)
# check_required_packages_and_commands ${required_pkgs[@]}
########
function check_required_packages_and_commands() {
  create_missing_packages_and_commands_list "$@"
  if [ -n "$MISSING_PKGS" ]; then
    error "The following packages and/or commands are not available:$MISSING_PKGS. Please install them and try again."
  fi
}


########
# Check whether a list of packages and/or commands are available and puts them in MISSING_PKGS env variable
# Usage:
# required_pkgs=(jq curl)
# create_missing_packages_and_commands_list ${required_pkgs[@]}
########
function create_missing_packages_and_commands_list() {
  export MISSING_PKGS
  local required_pkgs=("$@")
  local missing_pkgs=""
  for pkg in "${required_pkgs[@]}"; do
    local package_or_command_exists
    check_package_or_command "$pkg" && package_or_command_exists=0 || package_or_command_exists=1
    if [ $package_or_command_exists == 1 ]; then
      missing_pkgs+=" $pkg"
    fi
  done
  MISSING_PKGS="$missing_pkgs"
}

########
# Get value from config json
# Globals:
#   INFILE
# Arguments:
#   json_path
# Outputs:
#   Returns the value or exits if not found
########
function get_mandatory_json_value() {
  local json_path="$1"
  local value

  value="$(jq -r "${json_path}" < "${INFILE}")"
  if [[ ${value} != null ]]; then
    echo "${value}"
  else
    error "${json_path} not found in input file"
  fi
}

########
# Get value from inline json
# Arguments:
#   json_path
#   json_string
# Outputs:
#   Returns the value or exits if not found
########
function get_mandatory_inline_json_value() {
  local json_string="$2"
  local json_path="$1"
  local value
  #shellcheck disable=SC2015
  value="$(echo "${json_string}" | jq -r "${json_path}")" \
    && echo "${value}" \
    || error "${json_path} not found in input json"
}

########
# Get optional value from inline json
# Arguments:
#   json_path
#   json_string
# Outputs:
#   Returns the value or an empty string if not found
########
function get_optional_inline_json_value() {
  local json_string="$2"
  local json_path="$1"
  local value
  #shellcheck disable=SC2015
  value="$(echo "${json_string}" | jq -r "${json_path}")"
  if [[ ${value} != null ]]; then
    echo "${value}"
  fi
}

########
# Get version of chart from helm-charts.json
# Arguments:
#   chart name
# Outputs:
#   Returns the value of ".version" or an empty string if not found
########
function get_chart_version_from_manifest() {
  local chart_name="$1"
  local section="${2:-"uipath"}"
  local manifest="${MODULES_DIR}/../versions/helm-charts.json"
  local value
  #shellcheck disable=SC2015
  value="$(jq -r --arg CHARTNAME "${chart_name}" --arg SECTION "${section}_charts" '.[$SECTION] | .[] | select(.name==$CHARTNAME) | .version' < "${manifest}" )" \
    && echo "${value}" \
    || error "${chart_name} not found in ${manifest} under .${section}_charts."
}

########
# Get value from config json and base64 encode
# Globals:
#   INFILE
# Arguments:
#   json_path
# Outputs:
#   Returns the base64 encoded value or exits if not found
########
function get_mandatory_b64enc_json_value() {
  local json_path="$1"
  local value

  value="$(jq -r "${json_path}" < "${INFILE}")"
  if [[ ${value} != null ]]; then
    echo -n "${value}" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  else
    error "${json_path} not found in input file"
  fi
}

########
# Get value from config json
# Globals:
#   INFILE
# Arguments:
#   json_path
# Outputs:
#   Returns the value or exits if not found
########
function get_optional_json_value() {
  local json_path="$1"
  local value

  value="$(jq -r "${json_path}" < "${INFILE}")"
  if [[ ${value} != null ]]; then
    echo "${value}"
  fi
}

########
# Get value from config json and base64 encode
# Globals:
#   INFILE
# Arguments:
#   json_path
# Outputs:
#   Returns the base64 encoded value or exits if not found
########
function get_optional_b64enc_json_value() {
  local json_path="$1"
  local value

  value="$(jq -r "${json_path}" < "${INFILE}")"
  if [[ ${value} != null ]]; then
    echo -n "${value}" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  else
    echo "" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  fi
}


#####
# Open Port Forward to service or pod
# To use:
# openport argocd service/argocd-server 443
#####
function openport() {
    local try=0
    local maxtry=30
    local PORTFORWARD_PORT=""
    local NAMESPACE=$1
    local ENDPOINT=$2
    local PORT=$3
    kubectl port-forward -n "$NAMESPACE" "$ENDPOINT" ":$PORT" >>"${dev_null}" &
    while ((try != maxtry)) && [[ -z $PORTFORWARD_PORT ]] ; do
        try=$((try + 1))
        PORTFORWARD_PORT=$(ss -lp4 | grep "pid=$!" | awk '{print $5}' | awk -F ":" '{print $2}')
        sleep 1
    done
    if [[ -z $PORTFORWARD_PORT ]]; then
        error "Could not open port $PORT to $ENDPOINT in namespace $NAMESPACE"
    else
        echo "$PORTFORWARD_PORT"
    fi
}

#####
# Kill existing kubectl port-forward process based on port
# To use:
# closeport 31842
#####
function closeport() {
    local PORTFORWARD_PORT=$1
    local PID
    PID=$(ss -lp4 | grep "kubectl" | grep "127.0.0.1:$PORTFORWARD_PORT" | awk '{print $7}' | awk -F "pid=" '{print $2}' | awk -F "," '{print $1}')
    kill "$PID"
}

####
# Wait for job to become completed, filter by label
####
function wait_for_job_complete_by_label() {
  local label=$1
  local namespace=$2
  local try=0
  local maxtry=50
  local status="notready"
  while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
    try=$((try+1))
    kubectl -n "$namespace" wait --for=condition=complete job -l "$label" && status="ready"
    info "Waiting for the $label jobs to complete...${try}/${maxtry}" && sleep 30;
  done
  [[ ${status} == "ready" ]] || error "Failed to complete $label jobs."
}


########
# Install Packages depending on OS
########
function load_prerequisites_package_utils() {
  detect_os
  local MODULES_DIRECTORY
  MODULES_DIRECTORY=$1
  if [ "$OS_ID" == "rhel" ]; then
    #shellcheck disable=SC1090,SC1091
    source "$MODULES_DIRECTORY/install_packages/prerequisites-$OS_ID.sh"
  else
    warn "Prerequisites installation of packages is not supported for $OS_ID"
  fi
}

########
# Check if the script is getting triggered from root or not
########
function require_root_user() {
    local user=""
    user="$(id -un 2>>"${dev_null}" || true)"
    if [ "$user" != "root" ]; then
        error "Infra installer need to be run as root."
    fi
}

########
# Get value from config json or return default value
# Globals:
#   INFILE
# Arguments:
#   json_path
#   default_value
# Outputs:
#   Returns the value or return default value if not found
########
function get_json_value_or_return_default_value() {
  local json_path="$1"
  local default_value="$2"
  local value
  value="$(get_optional_json_value "${json_path}")"
  if [[ -n ${value} ]]; then
    echo "${value}"
  else
    echo "${default_value}"
  fi
}

function failure_output() {
  local title=$1
  shift
  local failures=("$@")
  printf '=%.0s' {1..100}
  echo
  error_without_exit "$title"
  for error in "${failures[@]}"
  do
    [[ -n ${error} ]] && error_without_exit ">>> $error"
  done
  printf '=%.0s' {1..100}
  echo
  exit 1
}

function warn_output() {
  local title=$1
  shift
  local warnings=("$@")
  printf '=%.0s' {1..100} >&2
  echo >&2
  warn "$title"
  for warn in "${warnings[@]}"
  do
    [[ -n ${warn} ]] && warn ">>> $warn"
  done
  printf '=%.0s' {1..100} >&2
  echo >&2
}

########
# Validates whether the argument is empty or not
########
function check_if_empty() {
    local arg_name=$1
    local arg_val=$2

    if [ -z "$arg_val" ]; then
        error "$arg_name argument is missing"
    fi
}


# Validate if any non core service is enabled
function is_full_installation(){
  local aicenter
  local business_apps
  local du
  local taskmining
  #aicenter
  aicenter=$(get_optional_json_value ".aicenter.enabled")
  #business-apps
  business_apps="$(get_optional_json_value ".apps.enabled")"
  #DU
  du="$(get_optional_json_value ".documentunderstanding.enabled")"
  #taskmining
  taskmining="$(get_optional_json_value ".task_mining.enabled")"

  if [[ ( -n "${aicenter}" && "${aicenter}" == "true" ) ||
    ( -n "${taskmining}" && "${taskmining}" == "true" ) ||
    ( -n "${du}" && "${du}" == "true" ) ||
    ( -n "${business_apps}" && "${business_apps}" == "true" ) ]]; then
    echo "true"
  else
    echo "false"
  fi
}

function get_existing_trap() {
  local signal=$1
  (( $# == 1 )) ||  error "${FUNCNAME[0]} needs a signal parameter"
  trap_cmd() { printf '%s\n' "$3"; }
  eval trap_cmd "$(trap -p "${signal}")"
}

# call option with parameters: $1=name $2=value $3=file
function add_or_replace_option() {
    name=${1//\//\\/}
    value=${2//\//\\/}
    sed -i \
        -e '/^#\?\(\s*'"${name}"'\s*=\s*\).*/{s//\1'"${value}"'/;:a;n;ba;q}' \
        -e '$a'"${name}"'='"${value}" "$3"
}

# It will fetch value associated with parameter for auditctl -s command
# Make sure you check `systemctl is-enabled auditd` before calling this function
function get_auditctl_parameter_value() {
  local parameter_name=$1
  local output
  output=$(auditctl -s | grep "${parameter_name}" | tail -1)
  IFS=' ' read -r -a output_value <<<"${output}"
  echo "${output_value[1]}"
}

function stop_incoming_connections_istio() {
  # Disable incoming calls
  kubectl -n istio-system patch daemonset istio-ingressgateway -p '{"spec": {"template": {"spec": {"nodeSelector": {"non-existing": "true"}}}}}' &>>"${dev_null}" || true
}

function start_incoming_connections_istio() {
  # enable gateway, as we need it during upgrade
  kubectl -n istio-system patch daemonset istio-ingressgateway --type json -p='[{"op": "remove", "path": "/spec/template/spec/nodeSelector/non-existing"}]' &>>"${dev_null}" || true
}

function wait_for_istio_pods() {
  kubectl -n istio-system rollout status daemonset istio-ingressgateway
}


export BACKUP_MARKER_FILE="/etc/rancher/as-backup-marker.txt"
export CONSENT_FILE="/etc/rancher/as-upgrade-consent.txt"
export RECURRING_JOB_BACKUP_LOCATION="/var/lib/uipath/uipath-backup-recurringjob.yaml"
export BACKUP_CRONTAB_LOCATION="/var/lib/uipath/crontab.bakup"
export MAINTENANCE_MODE_DEPLOY_REPLICA_CM="uipath-deployment-replicas"

function delete_recurring_job() {
  backup_job_name="${1}"

  # validate pvc cronjob for backup created
  local try=0
  local maxtry=10
  local exists=""
  while (( try < maxtry ));
  do
    # delete recurring job
    kubectl -n longhorn-system delete recurringjob "${backup_job_name}" --force || true
    kubectl -n longhorn-system delete cronjobs "${backup_job_name}" --force || true

    exists=$(kubectl -n longhorn-system get cronjobs "${backup_job_name}" &>>"${dev_null}") || true
    if [[ -z "${exists}" ]]; then
      info "Recurring cronjob ${backup_job_name} deleted successfully"
      break;
    fi

    warn "Recurring cronjob ${backup_job_name} not deleted yet, ${try}/${maxtry}..."; sleep 5
    try=$((try + 1))
    exists=""
  done

  if [[ -n "${exists}" ]];then
    error "Failed to delete recurring job for ${backup_job_name}"
  fi
}

function update_uipath_backup(){
  local bool=$1
  info "updating flag backup_enabled=${bool} in uipath-infra/uipath-backup"
  bool_base64enc=$(echo "${bool}"|base64)
  if kubectl -n uipath-infra get secret uipath-backup &>/dev/null; then
    kubectl -n uipath-infra patch secret uipath-backup --type='json' -p="[{\"op\" : \"replace\" ,\"path\" : \"/data/backup_enabled\" ,\"value\" : \"${bool_base64enc}\"}]"
    info "backup_enabled=${bool} is updated in uipath-infra/uipath-backup"
  else
    warn "secret uipath-infra/uipath-backup not found"
  fi
}

function update_backup_ds() {
   local is_ds_patched=""
   local maxtry=5
   local try

   info "patch daemonset/backup-service. This step is required to disable backup daemonset and avoid copying new snapshot"
   if kubectl -n uipath-infra get daemonset backup-service ; then
     while [[ -z "${is_ds_patched}" ]] && (( try < maxtry )); do
       try=$(( try + 1 ))
       kubectl -n uipath-infra patch daemonset backup-service -p '{"spec": {"template": {"spec": {"nodeSelector": {"non-existing": "true"}}}}}'
       is_ds_patched=$(kubectl -n uipath-infra get ds backup-service --show-labels | grep "non-existing=true")
       if [[ -z "${is_ds_patched}" ]]; then
         info "patching daemonset/backup-service failed. Wait and retry ${try}/${maxtry}"; sleep 20
       fi
     done
     if [[ -z "${is_ds_patched}" ]]; then
       error "updating daemonset/backup-service failed."
     else
       info "daemonset/backup-service patched successfully"
     fi
   else
     info "daemonset backup-service is not present."
   fi

}

function update_backup_secret() {
  local backup_enabled="false"

  # disable backup-service
  update_uipath_backup "false"
  # check if flag value is updated
  if kubectl -n uipath-infra get secret uipath-backup ; then
    backup_enabled=$(kubectl -n uipath-infra get secret uipath-backup -o json| jq -r ".data.backup_enabled"|base64 -d)
  fi

  if [[ "${backup_enabled}" == "true" ]]; then
    error "Failed to disable backup in secret uipath-infra/uipath-backup."
  fi
}

function pause_backups_temp() {
  # Disable etcd backup
  info "Disabling regular backups for now."
  mkdir -p /var/lib/uipath/
  ([[ -f "${BACKUP_CRONTAB_LOCATION}" ]] && grep -q backupjob "${BACKUP_CRONTAB_LOCATION}") || crontab -l > ${BACKUP_CRONTAB_LOCATION}
  crontab -l > backupcron
  sed -i '/backupjob/d' backupcron
  crontab backupcron
  rm backupcron
 
  update_backup_ds
 
  info "start deleting longhorn recurring jobs"
  if kubectl -n longhorn-system get recurringjob uipath-backup &>>"${dev_null}";
  then
    kubectl -n longhorn-system get recurringjob uipath-backup -o yaml > "${RECURRING_JOB_BACKUP_LOCATION}"
    delete_recurring_job uipath-backup
    info "longhorn recurring jobs deleted"
  fi

  info "start disabling backup enabled flag"
  update_backup_secret

  info "Disabled regular backups during upgrade Successfully."
}

function resume_backup_schedule() {
  delete_tmp_files
  [[ -f "${RECURRING_JOB_BACKUP_LOCATION}" ]] && kubectl -n longhorn-system apply -f "${RECURRING_JOB_BACKUP_LOCATION}" && rm -f "${RECURRING_JOB_BACKUP_LOCATION}"
  [[ -f "${BACKUP_CRONTAB_LOCATION}" ]] && crontab "${BACKUP_CRONTAB_LOCATION}" && rm -f "${BACKUP_CRONTAB_LOCATION}"
  update_uipath_backup "true"
  if kubectl -n uipath-infra get daemonset backup-service &> /dev/null; then
    kubectl -n uipath-infra patch daemonset backup-service --type json -p='[{"op": "remove", "path": "/spec/template/spec/nodeSelector/non-existing"}]'
  fi
  info "Resumed scheduled backups on the node."
}

function save_latest_etcd_snapshots() {
  export PATH=$PATH:/usr/local/bin/
  # Take 2 etcd snapshots, --name is the prefix used for snapshot
  rke2 etcd-snapshot --name etcd-snapshot-ondemand
  rke2 etcd-snapshot --name etcd-snapshot-ondemand
}

function is_null_or_empty() {
  local value="$1"
  local result="false"
  if [[ -z "${value}" || "${value}" == " " || "${value}" == "null" ]]; then
    result="true"
  fi
  echo "${result}"
}

# https://uipath.atlassian.net/browse/SERFAB-7785
# We have observed a behaviour where an on-demand backup is copying an empty file to the NFS server. We use this data during cluster rollback and
# rollback will fail if this file is empty. Since we don't have any data available here and we can't modify already running Unix cronjob,
# adding these extra duplicate steps to eliminate this scenario and abort the upgrade operation if the generated file is empty

function create_disaster_recovery_cluster_config() {
  local tmpClusterConfigFile="cluster_config_tmp.json"
  local rke2_token fixed_rke2_address fqdn docker_username docker_password profile tmp backupdir
  backupdir=$(grep BACKUP_DIR= /usr/bin/backupjob | cut -d '=' -f2)
  info "Start fetching cluster info from secret"

  tmp=$(mktemp)
  # reset current file
  echo "{}" > "${tmpClusterConfigFile}"
  
  rke2_token=$(get_cluster_config "rke_token")
  if [[ $(is_null_or_empty "${rke2_token}") == "true" ]]; then
    error "rke2 token can't be empty as cluster rollback required valid rke2"
  fi

  fixed_rke2_address=$(get_cluster_config "fixed_rke_address")
  if [[ $(is_null_or_empty "${fixed_rke2_address}") == "true" ]]; then
    error "fixed_rke2_address can't be empty as cluster rollback required fixed_rke2_address"
  fi
  
  fqdn=$(get_cluster_config "cluster_fqdn")
  if [[ $(is_null_or_empty "${fqdn}") == "true" ]]; then
    error "cluster_fqdn can't be empty as cluster rollback required cluster_fqdn"
  fi

  profile=$(get_cluster_config "profile")
  if [[ $(is_null_or_empty "${profile}") == "true" ]]; then
    error "profile can't be empty as cluster rollback required this value"
  fi

  docker_username=$(get_cluster_config "infra.docker_registry.username")
  docker_password=$(get_cluster_config "infra.docker_registry.password")
  # also check the old location used to store these
  if [[ $(is_null_or_empty "${docker_username}") == "true" || $(is_null_or_empty "${docker_password}") == "true" ]]; then
    docker_username=$(kubectl -n uipath get secrets docker-registry-credentials -o jsonpath='{.data.username}' --ignore-not-found | base64 -d -w 0) || true
    docker_password=$(kubectl -n uipath get secrets docker-registry-credentials -o jsonpath='{.data.password}' --ignore-not-found | base64 -d -w 0) || true
  fi
  
  local use_cilium
  use_cilium=$(grep -c cilium /etc/rancher/rke2/config.yaml || true)

  info "Start creating ${tmpClusterConfigFile} file"

  jq --arg docker_username "$docker_username" '.infra.docker_registry.username = $docker_username' "${tmpClusterConfigFile}" > "${tmp}" && mv "${tmp}" "${tmpClusterConfigFile}"
  jq --arg docker_password "$docker_password" '.infra.docker_registry.password = $docker_password' "${tmpClusterConfigFile}" > "${tmp}" && mv "${tmp}" "${tmpClusterConfigFile}"
  jq --arg fqdn "$fqdn" '.fqdn = $fqdn' "${tmpClusterConfigFile}" > "${tmp}" && mv "${tmp}" "${tmpClusterConfigFile}"
  jq --arg address "$fixed_rke2_address" '.fixed_rke_address = $address' "${tmpClusterConfigFile}" > "${tmp}" && mv "${tmp}" "${tmpClusterConfigFile}"
  jq --arg token "$rke2_token" '.rke_token = $token' "${tmpClusterConfigFile}" > "${tmp}" && mv "${tmp}" "${tmpClusterConfigFile}"
  jq --arg profile "$profile" '.profile = $profile' "${tmpClusterConfigFile}" > "${tmp}" && mv "${tmp}" "${tmpClusterConfigFile}"
  jq --arg use_cilium "$use_cilium" '.use_cilium = $use_cilium' "${tmpClusterConfigFile}" > "${tmp}" && mv "${tmp}" "${tmpClusterConfigFile}" 

  clusterConfigFile="${backupdir}/cluster_config.json"
  echo "Start updating config file ${clusterConfigFile} on nfs server"
  cp "${tmpClusterConfigFile}" "${clusterConfigFile}"
}

# We need to copy rancher encryption config json because of rancher has restore issue with version v1.21.4+rke2r2
# We should remove this logic once rancher version is upgraded
# https://uipath.atlassian.net/browse/SERFAB-5256
function save_latest_rancher_encryption_config() {
  info "Triggering rancher encryption file config backup"
  mkdir -p /mnt/backup/backup/cred
  cp -r /var/lib/rancher/rke2/server/cred/encryption-config.json "/mnt/backup/backup/cred/"
}

function trigger_on_demand_backup() {
  info "Triggering etcd backup"
  save_latest_etcd_snapshots
  [[ -f /usr/bin/backupjob ]] && (/usr/bin/backupjob || error "Failed to run backup job")
  create_disaster_recovery_cluster_config
  info "Triggering volumes backup"
  if kubectl -n longhorn-system get cronjob uipath-backup &>>"${dev_null}";
  then
    kubectl -n longhorn-system delete job uipath-backup-manual 2>>"${dev_null}" || true
    kubectl -n longhorn-system create job --from=cronjob/uipath-backup "uipath-backup-manual"
    # wait for jobs completion
    info "Waiting for longhorn backup job to get completed"
    kubectl -n longhorn-system wait --for=condition=complete job/uipath-backup-manual --timeout=1800s
    save_latest_rancher_encryption_config
  fi
  info "On demand backup completed"
}

function cluster_backup() {
  if [[ ! -f ${CONSENT_FILE} ]] && [[ "${UPGRADE_SILENT_MODE}" == "false" ]];
  then
    # If we have not taken consent yet
    if crontab -l 2>>"${dev_null}" | grep -q backupjob;
    then
      ## Backup is enabled in the Cluster, check for SQL Server Backup as well
      printf "\n"
      printf "Now the setup will back up the cluster.\n"
      printf "The backup is used to restore the cluster to a previous working state in case of an upgrade failure.\n"
      printf "Warning: It is highly recommended to back up all the SQL Databases as well. Without it the data can be inconsistent in the cluster in the event of a restore.\n"
      printf "\n\n"
      printf "Enter your choice (y to continue and q to quit the upgrade) [y/q]:"
      read -r input

      case "${input}" in
          y) printf "Continuing the upgrade process with backup\n";
            echo "Upgrade Consent" > ${CONSENT_FILE};;
          q) exit 0;;
          *) printf "\nInvalid choice\n";
              exit 0;;
      esac
    else
      printf "\n"
      printf "Warning: Backup is not enabled in the Cluster.\n"
      printf "It is highly recommended to have the cluster backed up before upgrade.\n"
      printf "Without a backup, the cluster cannot be restored to a previous working state in case of an upgrade failure.\n"
      printf "\n\n"
      printf "Enter your choice (y to continue and q to quit the upgrade) [y/q]: "
      read -r input

      case "${input}" in
          y) printf "Continuing the upgrade process without backup\n";
            echo "Upgrade Consent" > ${CONSENT_FILE};;
          q) exit 0;;
          *) printf "\nInvalid choice\n";
              exit 0;;
      esac
    fi
  fi

  if [[ ! -f "${BACKUP_MARKER_FILE}" ]] && crontab -l | grep -q backupjob;
  then
    ## Trigger backup of the cluster.
    trigger_on_demand_backup
    ## Pause further backups
    pause_backups_temp
    ## Create a file to indicate Backup was completed
    echo "Backup Complete" > ${BACKUP_MARKER_FILE}
  fi
}

function delete_tmp_files(){
  rm -f ${BACKUP_MARKER_FILE}
  rm -f ${CONSENT_FILE}
}

function argocd_login_internal() {
  local BASE_DIR
  BASE_DIR=$(cd "${MODULES_DIR}/../Fabric_Installer" && pwd)
  [[ -n "${BINARY_DIR:-}" ]] || get_fabric_install_paths "${BASE_DIR}"
  [[ -n "${ARGOCD_NAMESPACE:-}" ]] || get_argocd_install_params &>>"${dev_null}"
  ARGOCD_PORT=443
  ARGOCD_FQDN=$(kubectl -n argocd get svc argocd-server -o jsonpath="{.spec.clusterIP}")
  ARGOCD_PASSWORD_SECRET="argocd-admin-password"
  export ARGO_PLAINTEXT="true"
  (login_argocd "${ARGOCD_PORT}" "${ARGOCD_FQDN}" "${ARGOCD_PASSWORD_SECRET}" "argocd" "argocd" &>> "${dev_null}") || true
  unset ARGO_PLAINTEXT
  argocd app list &>>"${dev_null}" || info "Could not log in to ArgoCD"
}

function update_configmap_with_replicas(){
  echo "$1" > deploy-replicas.json
  kubectl -n uipath create cm "${MAINTENANCE_MODE_DEPLOY_REPLICA_CM}" --from-file=deploy-replicas.json --dry-run=client -o yaml | kubectl apply -f -
}

function get_stored_replicas(){
  if kubectl -n uipath get cm "${MAINTENANCE_MODE_DEPLOY_REPLICA_CM}" &>>"${dev_null}";
  then
    kubectl -n uipath get cm "${MAINTENANCE_MODE_DEPLOY_REPLICA_CM}" -o jsonpath="{.data.deploy-replicas\.json}"
  else
    echo "[]"
  fi
}

function enable_maintenance_mode() {
  info "Putting the cluster in maintenance mode. This will take some time."

  WAIT_TIME="5m"
  if [[ -n "${1:-}" && "$1" == "wait" ]];
  then
    [[ -z "$2" ]] || WAIT_TIME="$2"
  fi

  sleep "$WAIT_TIME"
  # Scale down all uipath deployments
  UIPATH_DEPLOYMENTS=$(kubectl -n uipath get deploy --no-headers -o custom-columns=NAME:.metadata.name)
  ORIGINAL_REPLICAS=$(get_stored_replicas)
  for deploy in $UIPATH_DEPLOYMENTS;
  do
    stored_replica=$(echo "$ORIGINAL_REPLICAS" | jq --arg deploy "${deploy}" '.[] | select(.name==$deploy)' | jq '.replicas' | tr -d '"')
    # If there exists already an entry with this deployment's replica count, skip updating it
    if [[ -z "${stored_replica}" ]];
    then
      orig_replica=$(kubectl -n uipath get deploy "$deploy" -o jsonpath="{.spec.replicas}")
      ORIGINAL_REPLICAS=$(jq --arg replica "${orig_replica}" --arg deploy "${deploy}" '. += [{ name: $deploy, replicas : $replica }]' <<< "${ORIGINAL_REPLICAS}")
      update_configmap_with_replicas "${ORIGINAL_REPLICAS}"
    fi
    kubectl -n uipath scale deploy "$deploy" --replicas=0
    kubectl -n uipath rollout status deploy "$deploy" || true
  done

  terminatingpodscount=-1
  try=0
  maxtry=30
  while(( try < maxtry )); do
    try=$(( try + 1 ))
    terminatingpodscount=$(kubectl -n uipath get pods --no-headers | grep -c "Terminating") || true
    if [[ "${terminatingpodscount}" -le 1 ]]; then
      info "All pods in uipath namespace terminated"
      break
    fi
    info "Total pods in Terminating state ${terminatingpodscount}, Waiting for all pods to terminate... ${try}/${maxtry}"; sleep 10
  done

  # Suspend all cronjobs in uipath namespace
  UIPATH_CRONJOBS=$(kubectl -n uipath get cronjobs --no-headers -o custom-columns=NAME:.metadata.name)
  for cronjob in $UIPATH_CRONJOBS;
  do
    kubectl -n uipath patch cronjob "$cronjob" -p '{"spec" : {"suspend" : true }}'
  done

  # Add a label to the node indicating maintenance mode
  info "Adding maintenance label on the nodes"
  ALL_NODES=$(kubectl get nodes --no-headers -o custom-columns=NAME:.metadata.name)
  for node in $ALL_NODES ;
  do
    kubectl label --overwrite "nodes/$node" "uipath-maintenance=true" &>>"${dev_null}"
  done
}

function disable_maintenance_mode() {
  info "Removing cluster maintenance mode"
  [[ -n "${1:-}" && "$1" == "--no-argo-login" ]] || argocd_login_internal

  # Scale UP all uipath deployments
  UIPATH_DEPLOYMENTS=$(kubectl -n uipath get deploy --no-headers -o custom-columns=NAME:.metadata.name)
  ORIGINAL_REPLICAS=$(get_stored_replicas)
  if [[ "$ORIGINAL_REPLICAS" != "[]" ]];
  then
    for deploy in $UIPATH_DEPLOYMENTS;
    do
      orig_replica=$(echo "$ORIGINAL_REPLICAS" | jq --arg deploy "${deploy}" '.[] | select(.name==$deploy)' | jq '.replicas' | tr -d '"')
      if [[ -z "${orig_replica}" ]];
      then
        # If stored replica not found, set it to 1
        orig_replica=1
      fi
      kubectl -n uipath scale deploy "$deploy" --replicas="${orig_replica}"
    done
  fi
  kubectl -n uipath delete cm "${MAINTENANCE_MODE_DEPLOY_REPLICA_CM}" --ignore-not-found

  # Resume all cronjobs in uipath namespace
  UIPATH_CRONJOBS=$(kubectl -n uipath get cronjobs --no-headers -o custom-columns=NAME:.metadata.name)
  for cronjob in $UIPATH_CRONJOBS;
  do
    kubectl -n uipath patch cronjob "$cronjob" -p '{"spec" : {"suspend" : false }}'
  done

  ALL_APPS=$(argocd app list -o name 2>>"${dev_null}") || true
  for app in $ALL_APPS;
  do
    info "Syncing deployments in app ${app}"
    app_deployments=$(argocd app resources "$app" 2>>"${dev_null}" | grep Deployment | awk '{print $4}') || true
    for deploy in $app_deployments;
    do
      argocd app sync "$app" --resource "apps:Deployment:$deploy" --force &>>"${dev_null}" || true
    done

    info "Syncing cronjobs in app ${app}"
    app_cronjobs=$(argocd app resources "$app" 2>>"${dev_null}" | grep CronJob | awk '{print $4}') || true
    for cronjob in $app_cronjobs;
    do
      argocd app sync "$app" --resource "batch:CronJob:$cronjob" --force &>>"${dev_null}" || true
    done
  done

  # Scale down all uipath deployments
  UIPATH_DEPLOYMENTS=$(kubectl -n uipath get deploy --no-headers -o custom-columns=NAME:.metadata.name)
  for deploy in $UIPATH_DEPLOYMENTS;
  do
    replicas=$(kubectl -n uipath get deploy "${deploy}" -o jsonpath="{.spec.replicas}")
    if [[ $replicas == 0 ]];
    then
      warn "Could not scale back up $deploy deployments"
    fi
  done

  info "Removing maintenance label from the nodes"
  ALL_NODES=$(kubectl get nodes --no-headers -o custom-columns=NAME:.metadata.name)
  for node in $ALL_NODES ;
  do
    kubectl label --overwrite "nodes/$node" "uipath-maintenance"- &>>"${dev_null}"
  done
}

function is_maintenance_mode_enabled() {
  MAINTENANCE_NODES_COUNT=$(kubectl get nodes -o name --selector="uipath-maintenance=true" | wc -l)
  ALL_NODES_COUNT=$(kubectl get nodes -o name | wc -l)
  if [[ "${MAINTENANCE_NODES_COUNT}" == "${ALL_NODES_COUNT}" ]];
  then
    info "Yes. Maintenance mode is enabled on the Cluster"
  else
    info "No. Maintenance mode is not enabled in the Cluster"
  fi
}


function get_node_annotation_value() {
  local NODE_NAME
  local ANNOTATION_NAME
  NODE_NAME=$1
  ANNOTATION_NAME=$2
  VALUE=$(kubectl get nodes/"${NODE_NAME}" -o json | jq -er --arg ANNOTATION "${ANNOTATION_NAME}" '.metadata.annotations[$ANNOTATION]')
  [[ ${VALUE} == null ]] || echo "${VALUE}"
}

function update_node_annotation_value() {
  local NODE_NAME
  local ANNOTATION_NAME
  local ANNOTATION_VALUE
  NODE_NAME=$1
  ANNOTATION_NAME=$2
  ANNOTATION_VALUE=$3
  kubectl annotate --overwrite nodes/"${NODE_NAME}" "${ANNOTATION_NAME}=${ANNOTATION_VALUE}"
}

function add_resource_annotation_value() {
  local RESOURCE_NAME
  local ANNOTATION_NAME
  local ANNOTATION_VALUE
  RESOURCE_NAME=$1
  ANNOTATION_NAME=$2
  ANNOTATION_VALUE=$3
  kubectl annotate "${RESOURCE_NAME}" "${ANNOTATION_NAME}=${ANNOTATION_VALUE}" 2>>"${dev_null}" || info "${RESOURCE_NAME} already annotated; skipping..."
}

function determine_first_node_upgrade() {
  local kubectl_nodes
  local version
  local result
  version=$1
  result="true"
  [[ "$(systemctl is-enabled rke2-server 2>>"${dev_null}")" == "enabled" ]] || {
    # If it's an agent node, return false
    echo "false"
    return
  }
  kubectl_nodes=$(kubectl get nodes --no-headers)
  while read -r crt_node; do
    server_name=$(awk '{print $1}' <<<"${crt_node}")
    current_version=$(get_node_annotation_value "${server_name}" "uipath/infra-version")
    [[ "${current_version}" == "${version}" ]] && result="false"
  done <<<"${kubectl_nodes}"
  echo "${result}"
}

function get_enabled_services() {
  local enabled_sevices=()
  local service_iter
  local is_enabled
  ## Get this list from defaults.json
  local all_service_name=("platform" "orchestrator" "test_manager" "automation_hub" "automation_ops" "insights" "action_center" "aicenter" "apps" "documentunderstanding" "task_mining" "dataservice")

  if [[ ${INSTALLSTANDALONEAICENTER} == "1" ]] ; then
    enabled_sevices=("aicenter" "documentunderstanding")
  else
    for service_iter in "${all_service_name[@]}"
    do
      if [[ $service_iter == "platform" ]] || [[ $service_iter == "orchestrator" ]] ; then
        enabled_sevices+=("${service_iter}")
      else
        is_enabled="$(get_optional_json_value ".${service_iter}.enabled")"
        if [[ -n "${is_enabled}" ]] && [[ $is_enabled == "true" ]] ; then
          if [[ $service_iter == "test_manager" ]]; then
            enabled_sevices+=("testmanager")
          elif [[ $service_iter == "automation_hub" ]]; then
            enabled_sevices+=("automationhub")
          elif [[ $service_iter == "automation_ops" ]]; then
            enabled_sevices+=("automationops")
          elif [[ $service_iter == "action_center" ]]; then
            enabled_sevices+=("actioncenter")
          elif [[ $service_iter == "apps" ]]; then
            enabled_sevices+=("ba")
          elif [[ $service_iter == "task_mining" ]]; then
            enabled_sevices+=("taskmining")
          else
            enabled_sevices+=("${service_iter}")
          fi
        fi
      fi
    done
  fi
  echo "${enabled_sevices[@]}" | tr ' ' ','
}

function add_test_result() {
  local test_status="$1"
  local test_details="$2"
  local test_mitigation="$3"
  local test_color
  local color_reset='\e[0m'
  local redirection_device="/dev/stdout"

    case "$test_status" in
    PASSED)
      ## Green
      test_color='\u001b[32m'
      ;;
    WARN)
      ## yellow
      test_color='\u001b[33m'
      ;;
    FAILED)
      ## red
      test_color='\u001b[31m'
      redirection_device="/dev/stderr"
      export EXIT_CODE=51
      ;;
    INFO)
      ##
      test_color='\u001b[36m'
      ;;
    *)
      error "Unsupported test status: ${test_status} while calling add_test_result"
      ;;
  esac

  if [[ -n "${test_mitigation}" ]] ; then
    echo -e ">>> [${test_color}${test_status}${color_reset}] ${test_details}\n    ${test_mitigation}" >> "${redirection_device}"
  else
    echo -e ">>> [${test_color}${test_status}${color_reset}] ${test_details}"  >> "${redirection_device}"
  fi
}

function validate_ipaddress() {
  local ip="$1"
  local ipv4_regex='^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$'
  local ipv6_regex='^(?:[0-9a-fA-F]{0,4}:){1,7}[0-9a-fA-F]{0,4}$'
  local exitCode=0

  if [[ -n "${ip}" ]] ; then
    if ! echo "$ip" | grep -q -P "$ipv4_regex|$ipv6_regex" ; then
      error_without_exit "Invalid IP address: '${ip}' provided"
      exitCode=1
    fi
  else
    error_without_exit "IP address not provided"
    exitCode=1
  fi
  return "${exitCode}"
}

function get_nfs_settings() {
  set_longhorn_url
  local url="${LONGHORN_URL}/v1/settings/backup-target"
  local nfs_target
  nfs_target=$(curl "${url}" | jq '.value' | xargs)
  nfs_target=${nfs_target#nfs://}
  export SET_MOUNT_PATH=${nfs_target##*:}
  export SET_NFS_SERVER=${nfs_target%:"${SET_MOUNT_PATH}"}
}

function set_docker_credentials() {
  if kubectl -n docker-registry get deploy docker-registry &>>"${dev_null}";
  then
    if kubectl -n uipath get secrets docker-registry-credentials &>>"${dev_null}";
    then
      docker_username=$(kubectl -n uipath get secrets docker-registry-credentials -o jsonpath='{.data.username}' | base64 -d -w 0)
      docker_password=$(kubectl -n uipath get secrets docker-registry-credentials -o jsonpath='{.data.password}' | base64 -d -w 0)
      jq --arg docker_username "$docker_username" '.infra.docker_registry.username = $docker_username' "$1" > "$tmp" && mv "$tmp" "$1"
      jq --arg docker_password "$docker_password" '.infra.docker_registry.password = $docker_password' "$1" > "$tmp" && mv "$tmp" "$1"
    else
      warn "Could not fetch and update docker registry credentials, please refer docs"
    fi
  else
    info "Skipping docker registry creds as it is online environments"
  fi
}

function set_cluster_details() {
  fqdn=$(kubectl -n istio-system get gateway main-gateway -o jsonpath='{.spec.servers[?(@.port.name=="https")].hosts[0]}')
  fixed_rke2_address=$(grep -A 1 tls-san /etc/rancher/rke2/config.yaml | tail -1 | cut -d ' ' -f2)
  rke2_token=$(grep token /etc/rancher/rke2/config.yaml | tail -1 | cut -d ' ' -f2)

  jq --arg fqdn "$fqdn" '.fqdn = $fqdn' "$1" > "$tmp" && mv "$tmp" "$1"
  jq --arg address "$fixed_rke2_address" '.fixed_rke_address = $address' "$1" > "$tmp" && mv "$tmp" "$1"
  jq --arg token "$rke2_token" '.rke_token = $token' "$1" > "$tmp" && mv "$tmp" "$1"
}

function set_restore_settings_from_cluster() {
  hostname=$(fetch_hostname)
  backupdir=$(grep BACKUP_DIR= /usr/bin/backupjob | cut -d '=' -f2)
  backupdir=${backupdir#/mnt/backup/backup/}
  backuppath=${backupdir%/"${hostname}"}
  get_nfs_settings

  jq --arg backuppath "$backuppath" '.restore.etcdRestorePath = $backuppath' "${1}" > "$tmp" && mv "$tmp" "${1}"
  jq --arg nfsendpoint "$SET_NFS_SERVER" '.restore.nfs.endpoint = $nfsendpoint' "${1}" > "$tmp" && mv "$tmp" "${1}"
  jq --arg mountpath "$SET_MOUNT_PATH" '.restore.nfs.mountpath = $mountpath' "${1}" > "$tmp" && mv "$tmp" "${1}"
}

function generate_restore_json() {
  local fqdn
  local final_file=generated_restore.json
  tmp=$(mktemp)
  if kubectl -n uipath-infra get secret saved-restore-json &>>"${dev_null}";
  then
    info "Saved restore settings found"
    kubectl -n uipath-infra get secret saved-restore-json -o json | jq '.data."generated_restore.json"' | xargs | base64 -d > ${final_file}
  else
    \cp "${MODULES_DIR}/../sample-restore.json" ${final_file}
  fi

  set_cluster_details ${final_file}
  set_restore_settings_from_cluster ${final_file}
  set_docker_credentials ${final_file}
  info "Restore json generated at generated_restore.json, please verify once before using it"
}

function save_restore_json() {
  tmp=$(mktemp)
  local backupjson="$1"
  local final_file=generated_restore.json
  info "Creating restore.json using $backupjson and saving in the cluster"
  \cp "${MODULES_DIR}/../sample-restore.json" ${final_file}
  backuppath=$(jq '.backup.etcdBackupPath' "${backupjson}" | xargs)
  nfsendpoint=$(jq '.backup.nfs.endpoint' "${backupjson}" | xargs)
  mountpath=$(jq '.backup.nfs.mountpath' "${backupjson}" | xargs)

  jq --arg backuppath "$backuppath" '.restore.etcdRestorePath = $backuppath' ${final_file} > "$tmp" && mv "$tmp" ${final_file}
  jq --arg nfsendpoint "$nfsendpoint" '.restore.nfs.endpoint = $nfsendpoint' ${final_file} > "$tmp" && mv "$tmp" ${final_file}
  jq --arg mountpath "$mountpath" '.restore.nfs.mountpath = $mountpath' ${final_file} > "$tmp" && mv "$tmp" ${final_file}
  if kubectl -n uipath-infra get secret saved-restore-json &>>"${dev_null}";
  then
    kubectl -n uipath-infra delete secret saved-restore-json
  fi
  kubectl -n uipath-infra create secret generic saved-restore-json --from-file=${final_file}
}

function parse_fqdn_values() {
  CLUSTER_FQDN="$(get_mandatory_json_value ".fqdn")"
  ALTERNATIVE_FQDN="$(get_optional_json_value ".alternative_fqdn")"
  [[ -z $ALTERNATIVE_FQDN ]] && ALTERNATIVE_FQDN="${CLUSTER_FQDN}"
  export CLUSTER_FQDN ALTERNATIVE_FQDN

  UIPATH_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
  RANCHER_SUBDOMAIN="$(get_mandatory_json_value ".subdomain.rancher_server")"
  RANCHER_FQDN="${RANCHER_SUBDOMAIN}.${ALTERNATIVE_FQDN}"
  MONITORING_SUBDOMAIN="$(get_mandatory_json_value ".subdomain.rancher_server")"
  INSIGHTS_SUBDOMAIN="$(get_mandatory_json_value ".subdomain.looker")"
  LOOKER_SUBDOMAIN="$(get_mandatory_json_value ".subdomain.looker")"
  LOOKER_FQDN="${LOOKER_SUBDOMAIN}.${CLUSTER_FQDN}"
  export UIPATH_EXTERNAL_PORT RANCHER_SUBDOMAIN RANCHER_FQDN MONITORING_SUBDOMAIN INSIGHTS_SUBDOMAIN LOOKER_SUBDOMAIN LOOKER_FQDN

  OBJECTSTORE_SUBDOMAIN="$(get_mandatory_json_value ".subdomain.rook_ceph")"
  CEPH_FQDN="${OBJECTSTORE_SUBDOMAIN}.${CLUSTER_FQDN}"
  CEPH_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
  CEPH_REPLICATION_FQDN="${OBJECTSTORE_SUBDOMAIN}.${ALTERNATIVE_FQDN}"
  export OBJECTSTORE_SUBDOMAIN CEPH_FQDN CEPH_EXTERNAL_PORT CEPH_REPLICATION_FQDN

  DOCKER_REGISTRY_SUBDOMAIN="$(get_mandatory_json_value ".subdomain.docker_registry")"
  DOCKER_REGISTRY_FQDN="${DOCKER_REGISTRY_SUBDOMAIN}.${ALTERNATIVE_FQDN}"
  DOCKER_REGISTRY_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
  export DOCKER_REGISTRY_SUBDOMAIN DOCKER_REGISTRY_FQDN DOCKER_REGISTRY_EXTERNAL_PORT

  ARGO_SUBDOMAIN="$(get_mandatory_json_value ".subdomain.argo_cd")"
  ARGOCD_FQDN="${ARGO_SUBDOMAIN}.${ALTERNATIVE_FQDN}"
  ARGOCD_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
  export ARGO_SUBDOMAIN ARGOCD_FQDN ARGOCD_EXTERNAL_PORT

}

function parse_certificate_file_paths() {
  CERT_CA_FILE_PATH="$(get_optional_json_value ".server_certificate.ca_cert_file")"
  CERT_CA_FILE_PATH="${CERT_CA_FILE_PATH/#\~/$HOME}"

  CERT_TLS_FILE_PATH="$(get_optional_json_value ".server_certificate.tls_cert_file")"
  CERT_TLS_FILE_PATH="${CERT_TLS_FILE_PATH/#\~/$HOME}" #allow paths relative to the home directory

  CERT_KEY_FILE_PATH="$(get_optional_json_value ".server_certificate.tls_key_file")"
  CERT_KEY_FILE_PATH="${CERT_KEY_FILE_PATH/#\~/$HOME}"

  ALT_CERT_CA_FILE_PATH="$(get_optional_json_value ".alternative_certificate.ca_cert_file")"
  ALT_CERT_CA_FILE_PATH="${ALT_CERT_CA_FILE_PATH/#\~/$HOME}"

  ALT_CERT_TLS_FILE_PATH="$(get_optional_json_value ".alternative_certificate.tls_cert_file")"
  ALT_CERT_TLS_FILE_PATH="${ALT_CERT_TLS_FILE_PATH/#\~/$HOME}" #allow paths relative to the home directory

  ALT_CERT_KEY_FILE_PATH="$(get_optional_json_value ".alternative_certificate.tls_key_file")"
  ALT_CERT_KEY_FILE_PATH="${ALT_CERT_KEY_FILE_PATH/#\~/$HOME}"

  export CERT_CA_FILE_PATH CERT_TLS_FILE_PATH CERT_KEY_FILE_PATH ALT_CERT_CA_FILE_PATH ALT_CERT_TLS_FILE_PATH ALT_CERT_KEY_FILE_PATH
}

function parse_token_signing_certs_path() {
  TOKEN_SIGNING_CERT="$(get_optional_json_value ".identity_certificate.token_signing_cert_file")"
  TOKEN_SIGNING_CERT="${TOKEN_SIGNING_CERT/#\~/$HOME}"

  TOKEN_SIGNING_PASS="$(get_optional_json_value ".identity_certificate.token_signing_cert_pass")"
  export TOKEN_SIGNING_CERT TOKEN_SIGNING_PASS
}

function print_center() {
  termwidth="120"
  padding="$(printf '%0.1s' ={1..500})"
  printf '%*.*s%s%*.*s\n' 0 "$(((termwidth-2-${#1})/2))" "$padding" "$1" 0 "$(((termwidth-1-${#1})/2))" "$padding"
}

function substitute_positional_value() {
  local message="$1"
  local values=("${@:2}")
  local position=0
  #shellcheck disable=SC2048
  for value in "${values[@]}"; do
    message=${message//\{$position\}/${value}}
    position=$((position+1))
  done
  echo "${message}"
}

function get_readiness_message() {
  local key="$1"
  local jsonFile="$2"
  local values=("${@:3}")
  local message
  message=$(jq ". | ${key}.message" "${jsonFile}" | xargs)
  if [[ "${message}" == "null" ]]; then
    message=""
  fi
  substitute_positional_value "${message}" "${values[@]}"
}

function get_readiness_mitigation() {
  local key="$1"
  local jsonFile="$2"
  local values=("${@:3}")
  local mitigationMessage mitigationArr mitigationDelimiter

  mitigationMessage=""
  mitigationArr=$(jq -r ". | ${key}.mitigation" "${jsonFile}")
  if [[ "${mitigationArr}" != "null" ]]; then
    while read -r mitigation; do
      if [[ -n ${mitigation} ]]; then
        mitigation=$(sed -e 's/^"//' -e 's/"$//' <<<"${mitigation}")
        mitigationMessage+="${mitigationDelimiter}${mitigation}"
        mitigationDelimiter="\n    "
      fi
    done <<<"$(echo "${mitigationArr}" | jq ".[]")"
  fi
  substitute_positional_value "${mitigationMessage}" "${values[@]}"
}

function get_mitigation_string_from_array() {
  messages=("$@")
  printf -v return_message "%s\n    " "${messages[@]}"
  return_message=${return_message%?????}
  echo "$return_message"
}

########
# Label configMap/secret to enable secret/configmap copy to other namespaces
# Arguments:
#   namespace
#   object_type
#   object_name
########
function label_object() {
    local namespace=$1
    local object_type=$2
    local object_name=$3

    check_if_empty namespace "$namespace"
    check_if_empty object_type "$object_type"
    check_if_empty object_name "$object_name"

    kubectl -n "$namespace" label "$object_type" "$object_name" config-discovery=yes 2>>"${dev_null}" || true
}

function get_pvc_size() {
  local component="$1"
  local pvc_selector
  local pvc_ns
  local max_pvc_size
  local pvc_selector_prefix="-l"

  case "$component" in
    alertmanager)
      pvc_selector="app.kubernetes.io/name=alertmanager"
      pvc_ns="cattle-monitoring-system"
      ;;
    prometheus)
      pvc_selector="app.kubernetes.io/name=prometheus"
      pvc_ns="cattle-monitoring-system"
      ;;
    rabbitmq)
      pvc_selector="app.kubernetes.io/name=rabbitmq"
      pvc_ns="rabbitmq"
      ;;
    insights-lookerdir)
      pvc_selector_prefix=""
      pvc_selector="insights-looker-lookerdir-pvc"
      pvc_ns="uipath"
      ;;
    insights-datadir)
      pvc_selector_prefix=""
      pvc_selector="insights-looker-datadir-pvc"
      pvc_ns="uipath"
      ;;
    mongodb-data)
      pvc_selector="volume-type=data"
      pvc_ns="mongodb"
      ;;
    mongodb-logs)
      pvc_selector="volume-type=logs"
      pvc_ns="mongodb"
      ;;
    *)
      error "Unsupported component: '${component}'"
      ;;
  esac

  max_pvc_size=$(kubectl -n $pvc_ns get pvc "${pvc_selector_prefix}" "${pvc_selector}" -o json 2>>"${dev_null}" | jq -r '.items[] | .status.capacity.storage' | numfmt --from=iec-i | sort -n  | tail -n 1 | numfmt --to=iec-i)
  echo "${max_pvc_size}"
}

function get_argocd_application_config_value() {
  local argocd_namespace="argocd"
  local application_name="$1"
  local config_path="$2"
  local config_value

  config_value=$(kubectl -n "${argocd_namespace}" get application "${application_name}" -o json 2>/dev/null | jq -r "${config_path}")
  echo "${config_value}"
}

function wait_till_cephobjectstore_ready() {
  local ceph_objectstore_name="$1"
  local ceph_objectstore_namespace="$2"
  local ceph_pod_labels="$3"
  local maxtry=120
  local try
  local ceph_objectstore_status
  local is_failure="true"
  local expected_ceph_objectstore_status="Connected"

  [ -z "$ceph_objectstore_name" ] && error "ceph_objectstore_name argument is missing for ${FUNCNAME[0]}"
  [ -z "$ceph_objectstore_namespace" ] && error "ceph_objectstore_namespace argument is missing for ${FUNCNAME[0]}"
  [ -z "$ceph_pod_labels" ] && error "ceph_pod_labels argument is missing for ${FUNCNAME[0]}"

  for try in $(seq 1 "${maxtry}") ; do
    ceph_objectstore_status=$(kubectl -n "${ceph_objectstore_namespace}" get cephobjectstore.ceph.rook.io "${ceph_objectstore_name}" -o json | jq -r '.status.phase')
    [[ "$ceph_objectstore_status" == "$expected_ceph_objectstore_status" ]] && is_failure="false" && break
    [[ $try -eq $maxtry ]] && break
    info "Waiting for cephobjectstore ${ceph_objectstore_name} to be ready. [Expected: $expected_ceph_objectstore_status, Actual: ${ceph_objectstore_status}]  (${try}/${maxtry})"
    sleep 5
  done
  [[ $is_failure == "true" ]] && error "Max try reach waiting for cephobjectstore: ${ceph_objectstore_name} to be ready. [Expected: $expected_ceph_objectstore_status, Actual: ${ceph_objectstore_status}] (${try}/${maxtry})"
  get_ready_pod "$ceph_objectstore_namespace" "$ceph_pod_labels" "300" >> /dev/null
  info "Cephobjectstore: ${ceph_objectstore_name} is ready"
}

function get_ready_pod() {
  local namespace="$1"
  local app_label="$2"
  local try
  local maxtry="${3:-120}"
  local ready_pod
  local is_failure="true"

  [ -z "$namespace" ] && error "namespace argument is missing for ${FUNCNAME[0]}"
  [ -z "$app_label" ] && error "app_label argument is missing for ${FUNCNAME[0]}"

  for try in $(seq 1 "${maxtry}") ; do
    ready_pod=$(kubectl -n "$namespace" get pods --field-selector=status.phase=Running -l "$app_label" -o json |
      jq -r '[.items[] | select((try (.status.conditions[] | select(.type=="Ready").status) // "False") == "True") | .metadata.name] as $podArray| if ($podArray | length) > 0 then $podArray[0] else "" end'
    )
    [[ -n "${ready_pod}" ]] && is_failure="false" && break
    [[ $try -eq $maxtry ]] && break
    info "Waiting for at-least one pod with label: $app_label  in namespace: ${namespace} to be ready (${try}/${maxtry})" >> /dev/stderr
    sleep 5
  done
  [[ $is_failure == "true" ]] && error "Max try reach waiting for at-least one pod with label: $app_label in namespace: ${namespace} to be ready (${try}/${maxtry})"
  printf "%s" "$ready_pod"
}

function is_ceph_osd_on_pv() {
  local return_code=1
  if PATH="$PATH:/var/lib/rancher/rke2/bin/" kubectl -n rook-ceph get pvc -l ceph.rook.io/DeviceSet=set1 -o name 2> /dev/null | grep -q 'persistentvolumeclaim'; then
    return_code=0
  fi
  return "${return_code}"
}

function is_ceph_pg_active_clean() {
  local return_code=1
  if kubectl -n rook-ceph exec  deploy/rook-ceph-tools -- ceph status --format json | jq '. as $root | ($root | .pgmap.num_pgs) as $total_pgs | try ( ($root | .pgmap.pgs_by_state[] | select(.state_name == "active+clean").count)  // 0) as $active_pgs | if $total_pgs == $active_pgs then true else false end' | grep -q 'true';then
    return_code=0
  fi
  return "${return_code}"
}

## This function modifies config for already running cluster
function speed_up_ceph_recovery(){
  local rook_ceph_namespace="rook-ceph"

  kubectl -n "${rook_ceph_namespace}" exec deploy/rook-ceph-tools -- ceph tell osd.* injectargs --osd-recovery-sleep-hdd=0.0 > /dev/null || true
  kubectl -n "${rook_ceph_namespace}" exec deploy/rook-ceph-tools -- ceph tell osd.* injectargs --osd-max-backfills=8 > /dev/null || true
  kubectl -n "${rook_ceph_namespace}" exec deploy/rook-ceph-tools -- ceph tell osd.* injectargs --osd-recovery-max-active-hdd=8 > /dev/null || true
  kubectl -n "${rook_ceph_namespace}" exec deploy/rook-ceph-tools -- ceph tell osd.* injectargs --osd-recovery-max-single-start=8 > /dev/null || true
}

function is_ceph_pool_exists() {
    local pool_name="$1"
    local rook_ceph_namespace="rook-ceph"
    local exit_code=1
    if kubectl -n "${rook_ceph_namespace}" exec deploy/rook-ceph-tools -- ceph osd pool ls  --format json | jq -r --arg poolname "${pool_name}" 'index($poolname) != null' | grep -q 'true'; then
        exit_code=0
    fi
    return "${exit_code}"
}

function get_ceph_pool_config() {
    local pool_name="$1"
    local pool_config="$2"
    local config_value
    local rook_ceph_namespace="rook-ceph"

    config_value=$(kubectl -n "${rook_ceph_namespace}" exec deploy/rook-ceph-tools -- ceph osd pool get "${pool_name}" "${pool_config}" --format json  | jq -r ".${pool_config}") || true
    echo "${config_value}"
}

function set_ceph_pool_config() {
    local pool_name="$1"
    local pool_config="$2"
    local config_value="$3"
    local rook_ceph_namespace="rook-ceph"

    kubectl -n "${rook_ceph_namespace}" exec deploy/rook-ceph-tools -- ceph osd pool set  "${pool_name}" "${pool_config}" "${config_value}" --yes-i-really-mean-it || true
}

function handle_device_health_metrics_pool_config() {
    if ! is_ceph_pool_exists "device_health_metrics"; then
        return
    fi
    rgw_data_pool_size=$(get_ceph_pool_config "rook-ceph.rgw.buckets.data" "size")
    [[ -z "${rgw_data_pool_size}" ]] && warn "Empty value for pool: 'rook-ceph.rgw.buckets.data' size" && return
    rgw_data_pool_min_size=$(get_ceph_pool_config "rook-ceph.rgw.buckets.data" "min_size")
    [[ -z "${rgw_data_pool_min_size}" ]] && warn "Empty value for pool: 'rook-ceph.rgw.buckets.data' min_size" && return

    set_ceph_pool_config "device_health_metrics" "size" "${rgw_data_pool_size}"
    set_ceph_pool_config "device_health_metrics" "min_size" "${rgw_data_pool_min_size}"
}

function wait_until_ceph_pgs_active() {
  local timeout_seconds=1200
  local start_time
  local current_time
  local diff
  local exit_code=0

  start_time=$(date +"%s")
  while ! is_ceph_pg_active_clean; do
    current_time=$(date +"%s")
    diff=$((current_time-start_time))
    if [[ $diff -gt $timeout_seconds ]]; then
      exit_code=1
      break
    fi
  done
  return "${exit_code}"
}

function is_label_value_valid() {
  local value="$1"
  echo "${value}" | grep -q -P '^(([A-Za-z0-9][-A-Za-z0-9_.]*)?[A-Za-z0-9])?$'
}
#####
# Check if server exists on machine
# Parameters:
#    service name (without the .service suffix)
#####
function service_exists() {
  local service_name=$1
  if [[ $(systemctl list-unit-files --no-legend "${service_name}.service" | cut -f1 -d' ') == "${service_name}.service" ]]; then
    return 0
  else
    return 1
  fi
}

function is_rke2_ambiguous() {
  if [[ $(service_exists "rke2-agent") && $(service_exists "rke2-server") ]]; then
    return 0
  else
    return 1
  fi
}

function get_overridden_value_from_json_for_external_storage() {
  local external_storage_json="$1"
  local service_specific_json="$2"
  local key="$3"

  local service_specific_value
  value="$(echo "${external_storage_json}" | jq -r ".${key}")"
  service_specific_value="$(echo "${service_specific_json}" | jq -r ".${key}")"

  [[ ${value} == null ]] && value=""
  [[ ${service_specific_value} != null ]] && value="${service_specific_value}"

  echo "${value}"
}

function get_overridden_bucket_name() {
  local create_bucket
  local external_storage_json="$1"
  local service_specific_json="$2"
  local prefix suffix default_bucket_name bucket_name

  create_bucket="$(get_mandatory_json_value ".external_storage.create_bucket")"
  default_bucket_name="$(echo "${service_specific_json}" | jq -r ".bucket_name")"
  prefix="$(get_overridden_value_from_json_for_external_storage "${external_storage_json}" "${service_specific_json}" "bucket_name_prefix")"
  suffix="$(get_overridden_value_from_json_for_external_storage "${external_storage_json}" "${service_specific_json}" "bucket_name_suffix")"
  bucket_name=""

  if [[ -z "${default_bucket_name}" ]]; then
    error "Bucket name can't be empty"
  fi

  if [[ "${create_bucket}" == "true" ]]; then
    [[ -n "${prefix}" ]] && bucket_name="${prefix}-"
    bucket_name="${bucket_name}${default_bucket_name}"
    [[ -n "${suffix}" ]] && bucket_name="${bucket_name}-${suffix}"
  else
    bucket_name="${default_bucket_name}"
  fi

  echo "${bucket_name}"
}

function get_local_k8s_node_name(){
  local node_name
  node_name=$(xargs -0 < "/proc/$(pgrep kubelet)/cmdline" | tr ' ' '\n' | grep '\-\-hostname\-override' | cut -d'=' -f2)
  [[ -z "${node_name}" ]] && node_name=$(fetch_hostname_lowercase)
  [[ -z "${node_name}" ]] && warn "Unable to get kubernetes node name"
  echo "${node_name}"
}

function rollout_restart_k8s_resource_and_wait() {
  local k8s_resource="$1"
  local namespace="$2"
  local timeout="${3:-300s}"

  kubectl -n "${namespace}" rollout restart "${k8s_resource}"
  kubectl -n "${namespace}" rollout status --timeout="${timeout}" "${k8s_resource}"
}

function add_empty_line_in_file_if_not_exists() {
  local file_name="$1"
  if [[ -n "$(tail -1 "${file_name}")" ]]; then
    echo "" >> "${file_name}"
  fi
}

# Check disk mounted at specific path
function is_partition_exist() {
  local dev_name=$1
  if lsblk  --noheadings --paths | grep -q -w "${dev_name}"; then
    echo "true"
  else
    echo "false"
  fi
}

