$code = @"
using System;
using System.Runtime.InteropServices;
    public static class GSSApiExtensions
    {
        private const string NetSecurityNative = "System.Net.Security.Native";
        private const string EnsureGssApiInitializedEntryPoint = "NetSecurityNative_EnsureGssInitialized";
        // This method is added to address the issue with NET6 regarding
        // Shim GSSAPI on Linux to delay loading libgssapi_krb5.so
        // SQL Client issue https://github.com/dotnet/SqlClient/issues/1390
        // SQL Client fix was merged in v4.x: https://github.com/dotnet/SqlClient/pull/1411#issuecomment-1009351682, but we want to avoid upgrading the library
        // dotnet runtime fix https://github.com/dotnet/runtime/pull/55037/files#diff-5cdec5583bd68ed1a0287de26e5bb8fc8e1ddc0981fc3b47ab515525cfc5bfe3R24
        public static void FixInitialization()
        {
            if (OperatingSystem.IsLinux() && Environment.Version.Major >= 6)
            {
                _ = EnsureGssInitialized();
            }
        }
        [DllImport(NetSecurityNative, EntryPoint = EnsureGssApiInitializedEntryPoint)]
        private static extern int EnsureGssInitialized();
    }
"@
Add-Type -TypeDefinition $code -Language CSharp
[GSSApiExtensions]::FixInitialization()
function Get-InputParameterPath {
    param(
        [Parameter(Mandatory)]
        [string] $Parameter
    )

    $filePath = "$env:SF_SECRETLOCATION/$Parameter"
    if (-Not(Test-Path -Path $filePath -PathType Leaf)) {
        throw "[$Parameter] Parameter does not exist"
    }

    return $filePath
}

function Get-InputParameters {
    param(
        [Parameter(Mandatory)]
        [string] $Parameter
    )

    $filePath = (Get-InputParameterPath -Parameter $Parameter)
    return (Get-Content -Path $filePath)
}

function Get-SanitizedConnectionString {
    param(
        [Parameter(Mandatory)]
        [string] $sqlConnectionString,

        [Parameter(Mandatory=$false)]
        [string] $ConnectionStringDriver = "System.Data.Common.DbConnectionStringBuilder"
    )

    $filePath = "$env:SF_BASEPATH/extraconfig/extra-config.json"
    if (-Not(Test-Path -Path $filePath -PathType Leaf)) {
        return $sqlConnectionString
    }
    $skipWords = (Get-Content -Path $filePath) | jq '.sqlConnectionStringUnsupportedKeys' | ConvertFrom-Json

    $connStringBuilder = New-Object $ConnectionStringDriver
    try {
        $connStringBuilder.set_ConnectionString($sqlConnectionString)
        foreach ($key in $skipWords) {
            $connStringBuilder.Remove("$key") | out-null
        }
    }
    catch {
        Write-Host "Error in getting sanitized connection string " $_
    }
    return $connStringBuilder.get_ConnectionString()
}

function Test-SqlHostPortConnectivity {
    param (
        [Parameter(Mandatory)]
        [string] $SqlServerHost,

        [Parameter(Mandatory)]
        [string] $SqlServerPort,

        [Parameter(Mandatory)]
        [string] $ServiceName,

        [Parameter(Mandatory)]
        [string] $TestName
    )

    $tcpClient = New-Object Net.Sockets.TcpClient
    try {
        $connectionResponse=$tcpclient.BeginConnect($SqlServerHost, $SqlServerPort, $null, $null)
        $connectionResponse.AsyncWaitHandle.WaitOne(500, $false)
        if( -Not $tcpclient.Connected ) {
            throw "Unable to reach sql sever on given fqdn and port."
        } else {
            $tcpclient.Close()
            Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "PASSED" -Message "Able to reach sql sever on given fqdn and port."
        }
    }
    catch {
        $errorMessage="$_"
        $errorMessage+="\n" + "* Please ensure, sql server firewall setting allowed to communicate with this machine"
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "FAILED" -Message "$errorMessage"
        return 1;
    }
    return 0;
}

function Test-DotNetDBConnection {
    param (
        [Parameter(Mandatory)]
        [string] $ConnectionStringParameter,

        [Parameter(Mandatory)]
        [string] $ServiceName,

        [Parameter(Mandatory)]
        [string] $TestName
    )
    $connectionString = Get-InputParameters -Parameter "$ConnectionStringParameter"
    $conn = New-Object System.Data.SqlClient.SqlConnection
    try {
        $conn.ConnectionString = (Get-SanitizedConnectionString -sqlConnectionString "$connectionString")
        $conn.Open()
        $message=[string]::format("Able to connect with {0} database using authentication mechanism provided", $ServiceName)
        Write-ValidationReport -Service $ServiceName -TestName $TestName -Level "PASSED" -Message $message
    } catch {
        $errorMessage="$_"
        if($errorMessage -like '*is not allowed to access the server*'){
            $errorMessage="Unable to reach sql sever on given fqdn and port."
            $errorMessage+="\n" + "$_"
        } else {
            $errorMessage=[string]::format("Unable to connect with {0} database using authentication mechanism provided.", $ServiceName)
            $errorMessage+="\n" + "$_"
        }
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "FAILED" -Message "$errorMessage"
        return 1;
    } finally {
        $conn.Close()
    }
    return 0;
}

function Test-DotNetDBConnection-FullTextSearchEnabled {
    param(
        [Parameter(Mandatory)]
        [string] $ConnectionStringParameter,

        [Parameter(Mandatory)]
        [string] $ServiceName,

        [Parameter(Mandatory)]
        [string] $TestName
    )
    $connectionString = Get-InputParameters -Parameter "$ConnectionStringParameter"
    $conn = New-Object System.Data.SqlClient.SqlConnection

    try {
        $conn.ConnectionString = (Get-SanitizedConnectionString -sqlConnectionString "$connectionString")
        $conn.Open()
        $Command = $conn.CreateCommand()
        $Command.CommandTimeout = 0
        $DataAdapter = New-Object System.Data.SqlClient.SqlDataAdapter $Command
        $Dataset = New-Object System.Data.Dataset
        $Command.CommandText = [string]::format("SELECT FULLTEXTSERVICEPROPERTY('IsFullTextInstalled')")
        $null = $DataAdapter.Fill($Dataset)

        if ($Dataset.Tables[0].Rows[0][0] -ne "1") {
            throw "FullTextSearch is not enabled"
        }

        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "PASSED" -Message "Full Text Search is enabled in sql sever. If you have enabled it, please make sure you have restarted deamon by running Exec sp_fulltext_service 'restart_all_fdhosts'"
    } catch {
        $errorMessage="$_"
        if($errorMessage -like '*FullTextSearch is not enabled*'){
            $errorMessage="Full Text Search is not enabled in sql sever."
            $errorMessage+="\n" + "$_"
        }
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "FAILED" -Message "$errorMessage"
        return 1;
    } finally {
        $conn.Close()
    }
    return 0;
}

function Test-ODBCDBConnection-FullTextSearchEnabled {
    param(
        [Parameter(Mandatory)]
        [string] $ConnectionStringParameter,

        [Parameter(Mandatory)]
        [string] $ServiceName,

        [Parameter(Mandatory)]
        [string] $TestName
    )
    $conn = New-Object System.Data.Odbc.OdbcConnection
    $connectionString = Get-InputParameters -Parameter "$ConnectionStringParameter"
    $conn.ConnectionString = (Get-SanitizedConnectionString -sqlConnectionString "$connectionString" -ConnectionStringDriver "System.Data.Odbc.OdbcConnectionStringBuilder")
    try {
        $conn.Open()

        $cmd = New-object System.Data.Odbc.OdbcCommand("SELECT FULLTEXTSERVICEPROPERTY('IsFullTextInstalled')", $conn)
        $Dataset = New-Object System.Data.DataSet
        (New-Object system.Data.odbc.odbcDataAdapter($cmd)).fill($Dataset) | out-null
        if ($Dataset.Tables[0].Rows[0][0] -ne "1") {
            throw "FullTextSearch is not enabled"
        }

        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "PASSED" -Message "Full Text Search is enabled in sql sever."

    } catch {
        $errorMessage="$_"
        if($errorMessage -like '*FullTextSearch is not enabled*'){
            $errorMessage="Full Text Search is not enabled in sql sever."
            $errorMessage+="\n" + "$_"
        }
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "FAILED" -Message "$errorMessage"
        return 1;
    } finally {
        $conn.Close()
    }
    return 0;
}

function Test-DotNetDBConnection-CompatibilityLevel {
    param(
        [Parameter(Mandatory)]
        [string] $ConnectionStringParameter,

        [Parameter(Mandatory)]
        [string] $ServiceName,

        [Parameter(Mandatory)]
        [string] $TestName,

        [Parameter(Mandatory=$false)]
        [bool] $CompatibilityLevelCheckOnly = $false
    )

    $Result = -100
    try {
        $conn = New-Object System.Data.SqlClient.SqlConnection
        $connectionString = Get-InputParameters -Parameter "$ConnectionStringParameter"
        $conn.ConnectionString = (Get-SanitizedConnectionString -sqlConnectionString "$connectionString")
        $conn.Open()

        $DbName = $conn.Database
        $Command = $conn.CreateCommand()
        $DataAdapter = new-object System.Data.SqlClient.SqlDataAdapter $Command
        $Dataset = new-object System.Data.Dataset

        if ($CompatibilityLevelCheckOnly -eq $true) {
            # Get compatibility level
            $Command.CommandText = "SELECT compatibility_level FROM sys.databases WHERE name = '${DbName}';"
            $null = $DataAdapter.Fill($Dataset)
            $Result = $Dataset.Tables[0].Item(0)
        } else {
            $TestTable = "TestTable"
            $Command.CommandText = "CREATE TABLE ${TestTable} (testID int);"
            $null = $DataAdapter.Fill($Dataset)
            try {
                # Check columnar is supported
                $Command.CommandText = "CREATE CLUSTERED COLUMNSTORE INDEX testIndex ON ${TestTable};"
                $null = $DataAdapter.Fill($Dataset)

                # Get compatibility level
                $Command.CommandText = "SELECT compatibility_level FROM sys.databases WHERE name = '${DbName}';"
                $null = $DataAdapter.Fill($Dataset)
                $Result = $Dataset.Tables[0].Item(0)

                # Check json is supported
                $Command.CommandText = @"
                    SELECT * FROM OpenJson(N'{
                        "String_value": "John",
                        "DoublePrecisionFloatingPoint_value": 45,
                        "DoublePrecisionFloatingPoint_value": 2.3456,
                        "BooleanTrue_value": true,
                        "BooleanFalse_value": false,
                        "Null_value": null,
                        "Array_value": ["a","r","r","a","y"],
                        "Object_value": {"obj":"ect"}
                    }');
"@
                $null = $DataAdapter.Fill($Dataset)
            } catch {
                $Result = -1
            } finally {
                $Command.CommandText = "DROP TABLE ${TestTable};"
                $null = $DataAdapter.Fill($Dataset)
            }
        }
    } catch {
        $Result = -999
    } finally {
        $conn.Close()
    }

    if ( $Result -eq -999 ) {
		Write-ValidationReport -Service "$ServiceName" -TestName "$TestName" -Level "FAILED" -Message "Unable to connect with database using authentication mechanism provided"
	} elseif ( 130 -gt $Result ) {
		Write-ValidationReport -Service "$ServiceName" -TestName "$TestName" -Level "FAILED" -Message "SQL server compatibility level must be >= 130, columnstore indexing must be supported, and json must be supported"
	} else {
		Write-ValidationReport -Service "$ServiceName" -TestName "$TestName" -Level "PASSED" -Message "SQL server compatibility level is >= 130 and Columnstore indexer, JSON support is enabled for Insights."
	}

    return $Result
}

function Test-ODBCConnection {
    param(
        [Parameter(Mandatory)]
        [string] $ConnectionStringParameter,

        [Parameter(Mandatory)]
        [string] $ServiceName,

        [Parameter(Mandatory)]
        [string] $TestName
    )
    try {
        $connectionString = Get-InputParameters -Parameter "$ConnectionStringParameter"
        $conn = New-Object System.Data.Odbc.OdbcConnection
        $conn.ConnectionString = (Get-SanitizedConnectionString -sqlConnectionString "$connectionString" -ConnectionStringDriver "System.Data.Odbc.OdbcConnectionStringBuilder")
        $conn.Open()
        $message=[string]::format("Able to connect with {0} database using authentication mechanism provided", $ServiceName)
        Write-ValidationReport -Service $ServiceName -TestName $TestName -Level "PASSED" -Message $message
    } catch {
        $errorMessage="$_"
        if($errorMessage -like '*is not allowed to access the server*'){
            $errorMessage="Unable to reach sql sever on given fqdn and port."
            $errorMessage+="\n" + "$_"
        } else {
            $errorMessage=[string]::format("Unable to connect with {0} database using authentication mechanism provided.", $ServiceName)
            $errorMessage+="\n" + "$_"
        }
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "FAILED" -Message "$errorMessage"
        return 1;
    } finally {
        $conn.Close()
    }
    return 0;
}

function Write-ValidationReport {
    param (
        [Parameter(Mandatory)]
        [string] $Service,

        [Parameter(Mandatory)]
        [string] $TestName,

        [Parameter(Mandatory)]
        [string] $Message,

        [Parameter(Mandatory)]
        [string] $Level,

        [string] $TestMitigation=""
    )

    $reportFile = "$env:VALIDATION_REPORT_FILE_LOCATION/report.json"

    If (-Not (Test-Path -Path $env:VALIDATION_REPORT_FILE_LOCATION -PathType Container)) {
        New-Item -Path $env:VALIDATION_REPORT_FILE_LOCATION -ItemType Directory
    }

    $object = New-Object PSCustomObject
    $object | Add-Member -Type NoteProperty -Name 'TestName' -Value $TestName
    $object | Add-Member -Type NoteProperty -Name 'Level' -Value $Level
    $object | Add-Member -Type NoteProperty -Name 'Message' -Value $Message
    $object | Add-Member -Type NoteProperty -Name 'TestMitigation' -Value $TestMitigation

    If (Test-Path -Path $reportFile -PathType Leaf) {
        $report = Get-Content -Path $reportFile | ConvertFrom-Json
        If($null -eq $report) {
            $report = New-Object PSCustomObject
        }
        $serviceReports = @($report.PSobject.Properties.Where( { $_.Name -eq $Service }))

        if ($serviceReports.value) {
            $found=0
            $serviceReports.value | ForEach-Object {
                $serviceReport=$_
                If (($serviceReport.TestName -eq $object.TestName) -And ($serviceReport.Level -eq $object.Level)) {
                    $found=1
                }
            }
            If ($found -eq 0) {
                $serviceReports = @($serviceReports.value) + @($object)
            } else {
                $serviceReports = @($serviceReports.value)
            }
            $report.psobject.properties.remove($Service)
        }
        else {
            $serviceReports = @($object)
        }
        $report | Add-Member -Type NoteProperty -Name $Service -Value $serviceReports
    }
    else {
        $report = New-Object PSCustomObject
        $report | Add-Member -Type NoteProperty -Name $Service -Value @($object)
    }

    $report | ConvertTo-Json -D 100 | Set-Content -Path $reportFile
}


function Write-ConsolidatedReport {
    $reportFile = "$env:VALIDATION_REPORT_FILE_LOCATION/report.json"

    $failed = $false
    If (Test-Path -Path $reportFile -PathType Leaf) {
        $reports = Get-Content -Path $reportFile | ConvertFrom-Json

        $services = @($reports.PSobject.Properties.Name)
        foreach ($service in $services) {

            Write-Host "`n###################################### $service ###########################################"
            $tests = $reports.PSobject.Properties.Where( { $_.Name -eq $service }).Value
            Write-Host ("{0, -50}{1, -10}{2}" -f "TestName", "Status", "Message")
            Write-Host "-------------------------------------------------------------------------------------------"
            foreach ($test in $tests) {
                Write-Host ("{0, -50}{1, -10}{2}`n" -f $test.TestName, $test.Level, $test.Message)

                $failed = $failed -or -Not($test.Level -eq "Passed")
            }
        }
    }
    return $failed
}


function Get-KerberosTgt {
    param(
        [Parameter(Mandatory)]
        [string] $ServiceName
    )

    if ($EnableIntegratedAuth -eq $true) {
        Write-Host "Integrated auth flag is $EnableIntegratedAuth"
        $ServiceNameUpper = $ServiceName.toUpper()
        $AdUser = -join($ServiceNameUpper, "_AD_USERNAME")
        $AdUserName = [System.Environment]::GetEnvironmentVariable($AdUser)
        if($AdUserName -and ($AdUserName -ne "PLACEHOLDER")){
            Write-Host "Service $ServiceName has specified AD account for Kerberos Auth. Execute: kinit $AdUserName -k"
            $KeyTab = -join($ServiceNameUpper, "_KRB5_KTNAME")
            $env:KRB5_KTNAME = [System.Environment]::GetEnvironmentVariable($KeyTab)
            kinit $AdUserName -k
        } else {
            Write-Host "Service $ServiceName is using default AD account for Kerberos Auth. Execute: kinit $env:AD_USERNAME -k"
            $env:KRB5_KTNAME = $env:DEFAULT_KRB5_KTNAME
            kinit $env:AD_USERNAME -k
        }
    }
}

function New-Database {
    param (
        [Parameter(Mandatory)]
        [string] $SqlDefaultConnectionParameter,

        [Parameter(Mandatory)]
        [string] $DBName,

        [Parameter(Mandatory)]
        [string] $ServiceName,

        [Parameter(Mandatory)]
        [string] $TestName
    )
    $connectionString = Get-InputParameters -Parameter "$SqlDefaultConnectionParameter"
    $conn = New-Object System.Data.SqlClient.SqlConnection

    try {
        $conn.ConnectionString = (Get-SanitizedConnectionString -sqlConnectionString "$connectionString")
        $conn.Open()
        $Command = $conn.CreateCommand()
        $Command.CommandTimeout = 0
        $DataAdapter = New-Object System.Data.SqlClient.SqlDataAdapter $Command
        $Dataset = New-Object System.Data.Dataset
        $Command.CommandText = [string]::format("IF NOT EXISTS ( SELECT name FROM master.dbo.sysdatabases WHERE name = '{0}' ) CREATE DATABASE {1};", $DBName, $DBName)
        $null = $DataAdapter.Fill($Dataset)
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "PASSED" -Message "SQL User have permissions to create database"
    } catch {
        $errorMessage="$_"
        if ($errorMessage -like '*CREATE DATABASE permission denied in database*') {
            $errorMessage="SQL User does not have permissions to create database."
            $errorMessage+="\n" + "$_"
        }
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "FAILED" -Message "$errorMessage"
        return 1;
    } finally {
        $conn.Close()
    }
    return 0;
}

function Remove-Database {
    param (
        [Parameter(Mandatory)]
        [string] $SqlDefaultConnectionParameter,

        [Parameter(Mandatory)]
        [string] $DBName
    )
    $connectionString = Get-InputParameters -Parameter "$SqlDefaultConnectionParameter"
    $conn = New-Object System.Data.SqlClient.SqlConnection

    try {
        $conn.ConnectionString = (Get-SanitizedConnectionString -sqlConnectionString "$connectionString")
        $conn.Open()
        $Command = $conn.CreateCommand()
        $Command.CommandTimeout = 0
        $DataAdapter = New-Object System.Data.SqlClient.SqlDataAdapter $Command
        $Dataset = New-Object System.Data.Dataset
        $Command.CommandText = [string]::format("DROP DATABASE IF EXISTS TestDb {0};", $DBName)
        $null = $DataAdapter.Fill($Dataset)
    } catch {
        return 1;
    } finally {
        $conn.Close()
    }
    return 0;
}

function New-TableAndPushTestData {
    param (
        [Parameter(Mandatory)]
        [string] $SqlDefaultConnectionParameter,

        [Parameter(Mandatory)]
        [string] $TableName,

        [Parameter(Mandatory)]
        [string] $ServiceName,

        [Parameter(Mandatory)]
        [string] $TestName
    )
    $connectionString = Get-InputParameters -Parameter "$SqlDefaultConnectionParameter"
    $conn = New-Object System.Data.SqlClient.SqlConnection
    try {
        $conn.ConnectionString = (Get-SanitizedConnectionString -sqlConnectionString "$connectionString")
        $conn.Open()
        $Command = $conn.CreateCommand()
        $Command.CommandTimeout = 0
        $DataAdapter = New-Object System.Data.SqlClient.SqlDataAdapter $Command
        $Dataset = New-Object System.Data.Dataset
        $Command.CommandText = [string]::format("CREATE TABLE {0} (testID int);", $TableName)
        $null = $DataAdapter.Fill($Dataset)

        $Command.CommandText = [string]::format("INSERT INTO {0} (testID) VALUES (1);", $TableName)
        $null = $DataAdapter.Fill($Dataset)

        $Command.CommandText = [string]::format("SELECT * FROM {0};", $TableName)
        $null = $DataAdapter.Fill($Dataset)

        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "PASSED" -Message "SQL User have permissions to create schema (db_ddladmin), read (db_datareader) and write (db_datawriter) to database."
    } catch {
        $errorMessage="SQL User have permissions to create schema (db_ddladmin), read (db_datareader) and write (db_datawriter) to database."
        $errorMessage+="\n" + "$_"
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "FAILED" -Message "$errorMessage"
        return 1;
    } finally {
        $Command.CommandText = "DROP TABLE $TableName;"
        $null = $DataAdapter.Fill($Dataset)
        $conn.Close()
    }
    return 0;
}

function Test-DotNetDatabaseRoles {
    param (
        [Parameter(Mandatory)]
        [string] $ConnectionStringParameter,

        [Parameter(Mandatory)]
        [string] $ServiceName,

        [Parameter(Mandatory)]
        [string] $TestName
    )
    $connectionString = Get-InputParameters -Parameter "$ConnectionStringParameter"
    $conn = New-Object System.Data.SqlClient.SqlConnection
    $Roles = @('db_ddladmin', 'db_datawriter', 'db_datareader')
    try {
        $conn.ConnectionString = (Get-SanitizedConnectionString -sqlConnectionString "$connectionString")
        $conn.Open()
        $SQLIsRoleMember = (Test-DotNetDatabaseRole -Connection $conn -RoleName "db_owner")
        if ($SQLIsRoleMember -ne 1) {
            $Roles | ForEach-Object {
                $SQLIsRoleMember = (Test-DotNetDatabaseRole -Connection $conn -RoleName $_)
                if ($SQLIsRoleMember -ne 1) {
                    throw "SQL User does not have permissions to either create schema (db_ddladmin) or read (db_datareader) or write (db_datawriter) to database"
                }
            }
        }
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "PASSED" -Message "SQL User have permissions to create schema (db_ddladmin), read (db_datareader) and write (db_datawriter) to database"
    } catch {
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "FAILED" -Message "$_"
    } finally {
        $conn.Close()
    }
}

function Test-ODBCDatabaseRoles {
    param (
        [Parameter(Mandatory)]
        [string] $ConnectionStringParameter,

        [Parameter(Mandatory)]
        [string] $ServiceName,

        [Parameter(Mandatory)]
        [string] $TestName
    )
    $conn = New-Object System.Data.Odbc.OdbcConnection
    $connectionString = Get-InputParameters -Parameter "$ConnectionStringParameter"
    $Roles = @('db_ddladmin', 'db_datawriter', 'db_datareader')
    try {
        $conn.ConnectionString = (Get-SanitizedConnectionString -sqlConnectionString "$connectionString" -ConnectionStringDriver "System.Data.Odbc.OdbcConnectionStringBuilder")
        $conn.Open()
        $SQLIsRoleMember = (Test-ODBCDatabaseRole -Connection $conn -RoleName "db_owner")
        if ($SQLIsRoleMember -ne 1) {
            $Roles | ForEach-Object {
                $SQLIsRoleMember = (Test-ODBCDatabaseRole -Connection $conn -RoleName $_)
                if ($SQLIsRoleMember -ne 1) {
                    throw "SQL User does not have permissions to either create schema (db_ddladmin) or read (db_datareader) or write (db_datawriter) to database"
                }
            }
        }
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "PASSED" -Message "SQL User have permissions to create schema (db_ddladmin), read (db_datareader) and write (db_datawriter) to database"
    } catch {
        Write-ValidationReport -Service "$ServiceName" -TestName $TestName -Level "FAILED" -Message "$_"
    } finally {
        $conn.Close()
    }
}

function Test-DotNetDatabaseRole {
    param (
        [Parameter(Mandatory)]
        [System.Data.SqlClient.SqlConnection] $Connection,

        [Parameter(Mandatory)]
        [string] $RoleName
    )
    $SqlQuery = [string]::format("SELECT IS_ROLEMEMBER('{0}') AS 'role';", $RoleName)
    $SqlCommand = New-Object System.Data.SqlClient.SqlCommand($SqlQuery, $Connection)
    $IsRoleMember = [String]$SqlCommand.ExecuteScalar()
    return $IsRoleMember
}

function Test-ODBCDatabaseRole {
    param (
        [Parameter(Mandatory)]
        [System.Data.Odbc.OdbcConnection] $Connection,

        [Parameter(Mandatory)]
        [string] $RoleName
    )
    $SqlQuery = [string]::format("SELECT IS_ROLEMEMBER('{0}') AS 'role';", $RoleName)
    $SqlCommand = New-Object System.Data.Odbc.OdbcCommand($SqlQuery, $Connection)
    $IsRoleMember = [String]$SqlCommand.ExecuteScalar()
    return $IsRoleMember
}