#!/bin/bash

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)

if [[ -z "${SCRIPT_DIR}" ]]; then
    echo "Script dir not present"
    exit 1
fi

# namespaces
TARGET_NAMESPACE="uipath"
SOURCE_NAMESPACE="test-config-discovery"

  
# test secret name 
TEST_SECRET1="test-secret1"

# test secret data
TEST_DATA_KEY1="test1"
TEST_SECRET_DATA="ZHVtbXk="
UPDATE_SECRET_DATA="ZHVtbXky"

# test configmap name
TEST_CONFIGMAP1="test-configmap1"

#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}/utils.sh"

function create_secret() {
  RESULT="failed"
  test_secret_json="{\"apiVersion\":\"v1\",\"data\":{\"${TEST_DATA_KEY1}\":\"${TEST_SECRET_DATA}\"},\"kind\":\"Secret\",\"metadata\":{\"name\":\"${TEST_SECRET1}\",\"labels\":{\"config-discovery\":\"yes\"}},\"type\":\"Opaque\"}"
  echo "${test_secret_json}"| kubectl create -n "${SOURCE_NAMESPACE}" -f -

  result=$(kubectl -n "${SOURCE_NAMESPACE}" get secrets "${TEST_SECRET1}")
  if [[ -z "${result}" ]]; then
    RESULT="failed"
    return 0
  fi
  RESULT="success"
}

function create_configmap() {
  RESULT="failed"
  test_configmap_json="{\"apiVersion\":\"v1\",\"data\":{\"${TEST_DATA_KEY1}\":\"${TEST_SECRET_DATA}\"},\"kind\":\"ConfigMap\",\"metadata\":{\"name\":\"${TEST_CONFIGMAP1}\",\"labels\":{\"config-discovery\":\"yes\"}}}"
  echo "${test_configmap_json}"| kubectl create -n "${SOURCE_NAMESPACE}" -f -

  result=$(kubectl -n "${SOURCE_NAMESPACE}" get configmaps "${TEST_CONFIGMAP1}")
  if [[ -z "${result}" ]]; then
    RESULT="failed"
    return 0
  fi
  RESULT="success"
}

function test_config_discovery_argocd() {
    local sf_core_config_discovery_health
    sf_core_config_discovery_health=$(kubectl -n argocd get Application sfcore -o json|jq -c ".status|select(.health.status==\"Healthy\").operationState.syncResult.resources[]|select(.name == \"config-discovery\")")
    
    if [[ -z "${sf_core_config_discovery_health}" ]]; then
      mark_test_failed 1525201 /tmp/test_results.json
      return 0
    fi

    config_discovery_status=$(echo "${sf_core_config_discovery_health}"|jq -r ".status")
    if [[ "${config_discovery_status}" == "Synced" ]]; then
      mark_test_passed 1525201 /tmp/test_results.json
    else
      mark_test_failed 1525201 /tmp/test_results.json
    fi
}

function test_copy_secret_configmap() {
    local secret_created configmap_created

    create_secret
    secret_created=${RESULT}
    create_configmap
    configmap_created=${RESULT}
    unset RESULT

    if [[ "${secret_created}" != "success" || "${configmap_created}" != "success" ]]; then
        mark_test_failed 1525204 /tmp/test_results.json
        return 0
    fi
    sleep 30

    local secret_result configmap_result
    # check secret and configmap in target namespace
    secret_result=$(kubectl -n "${TARGET_NAMESPACE}" get secrets "${TEST_SECRET1}")
    configmap_result=$(kubectl -n "${TARGET_NAMESPACE}" get configmaps "${TEST_CONFIGMAP1}")
    if [[ -z "${secret_result}" || -z "${configmap_result}" ]]; then
        mark_test_failed 1525204 /tmp/test_results.json
    else
        mark_test_passed 1525204 /tmp/test_results.json
    fi
}

function test_updating_secret_configmap_key_value() {
    # updating value
    secret_result=$(kubectl -n "${SOURCE_NAMESPACE}" patch secret "${TEST_SECRET1}" --type='json' -p="[{\"op\" : \"replace\" ,\"path\" : \"/data/${TEST_DATA_KEY1}\" ,\"value\" : \"${UPDATE_SECRET_DATA}\"}]")
    configmap_result=$(kubectl -n "${SOURCE_NAMESPACE}" patch configmap "${TEST_CONFIGMAP1}" --type='json' -p="[{\"op\" : \"replace\" ,\"path\" : \"/data/${TEST_DATA_KEY1}\" ,\"value\" : \"${UPDATE_SECRET_DATA}\"}]")
    if [[ -z "${secret_result}" || -z "${configmap_result}" ]]; then
        mark_test_failed 1525205 /tmp/test_results.json
        return 0
    fi

    sleep 30

    configmap_data=$(kubectl -n "${TARGET_NAMESPACE}" get configmap "${TEST_CONFIGMAP1}" -o json| jq -c ".data")
    secret_data=$(kubectl -n "${TARGET_NAMESPACE}" get secret "${TEST_SECRET1}" -o json| jq -c ".data")
    if [[ "${secret_data}" != *"${UPDATE_SECRET_DATA}"* || "${configmap_data}" != *"${UPDATE_SECRET_DATA}"* ]]; then
        mark_test_failed 1525205 /tmp/test_results.json
    else
        mark_test_passed 1525205 /tmp/test_results.json    
    fi
    
    # updating key
    secret_result=$(kubectl -n "${SOURCE_NAMESPACE}" patch secret "${TEST_SECRET1}" --type='json' -p="[{\"op\" : \"remove\" ,\"path\" : \"/data/${TEST_DATA_KEY1}\"}]")
    configmap_result=$(kubectl -n "${SOURCE_NAMESPACE}" patch configmap "${TEST_CONFIGMAP1}" --type='json' -p="[{\"op\" : \"remove\" ,\"path\" : \"/data/${TEST_DATA_KEY1}\"}]")
    if [[ -z "${secret_result}" || -z "${configmap_result}" ]]; then
        mark_test_failed 1525205 /tmp/test_results.json
        return 0
    fi

    sleep 30

    configmap_data=$(kubectl -n "${TARGET_NAMESPACE}" get configmap "${TEST_CONFIGMAP1}" -o json| jq -c ".data")
    secret_data=$(kubectl -n "${TARGET_NAMESPACE}" get secret "${TEST_SECRET1}" -o json| jq -c ".data")
    if [[ "${secret_data}" == *"${UPDATE_SECRET_DATA}"* || "${configmap_data}" == *"${UPDATE_SECRET_DATA}"* ]]; then
        mark_test_failed 1525207 /tmp/test_results.json
    else
        mark_test_passed 1525207 /tmp/test_results.json    
    fi
}

function test_deleting_secret_configmap() {
    kubectl -n "${SOURCE_NAMESPACE}" delete secret "${TEST_SECRET1}"
    kubectl -n "${SOURCE_NAMESPACE}" delete configmap "${TEST_CONFIGMAP1}"

    sleep 30

    secret_result=$(kubectl -n "${TARGET_NAMESPACE}" get secrets "${TEST_SECRET1}")
    configmap_result=$(kubectl -n "${TARGET_NAMESPACE}" get configmaps "${TEST_CONFIGMAP1}")

    if [[ -n "${secret_result}" || -n "${configmap_result}" ]]; then
        mark_test_failed 1525206 /tmp/test_results.json
    else
        mark_test_passed 1525206 /tmp/test_results.json
    fi    
}

function test_disable_enable_config_discovery_copies_secret_configmap() {
    kubectl -n uipath-infra scale deploy config-discovery --replicas=0
    sleep 5
    kubectl wait pod -n uipath-infra --selector deploy=config-discovery --for=delete --timeout=5m

    local secret_created configmap_created

    create_secret
    secret_created=${RESULT}
    create_configmap
    configmap_created=${RESULT}
    unset RESULT

    if [[ "${secret_created}" != "success" || "${configmap_created}" != "success" ]]; then
        mark_test_failed 1525208 /tmp/test_results.json
        return 0
    fi

    sleep 30

    kubectl -n uipath-infra scale deploy config-discovery --replicas=1
    sleep 5
    kubectl wait pod -n uipath-infra -l "deploy=config-discovery" --for=condition=Ready --timeout 2m

    sleep 60

    local secret_result configmap_result
    # check secret and configmap in target namespace
    secret_result=$(kubectl -n "${TARGET_NAMESPACE}" get secrets "${TEST_SECRET1}")
    configmap_result=$(kubectl -n "${TARGET_NAMESPACE}" get configmaps "${TEST_CONFIGMAP1}")
    if [[ -z "${secret_result}" || -z "${configmap_result}" ]]; then
        mark_test_failed 1525208 /tmp/test_results.json
    else
        mark_test_passed 1525208 /tmp/test_results.json
    fi
}

function main() {
    test_config_discovery_argocd
    # create namespace
    kubectl create namespace "${SOURCE_NAMESPACE}"

    test_copy_secret_configmap
    test_updating_secret_configmap_key_value
    test_deleting_secret_configmap
    test_disable_enable_config_discovery_copies_secret_configmap

    # delete namespace
    kubectl delete namespace "${SOURCE_NAMESPACE}"
}

main
