#!/bin/bash

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)

if [[ -z "${SCRIPT_DIR}" ]]; then
  echo "Script dir not present"
  exit 1
fi

if [[ ! -s "${SCRIPT_DIR}/utils.sh" ]]; then
  echo "Test utils.sh not found at ${SCRIPT_DIR}/utils.sh"
  exit 1
fi
#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}/utils.sh"

trap on_exit 0

function on_exit() {
  restore_resolver_file
  clean_up
}

function clean_up() {
  \rm "${DNS_RESOLVER_FILE}.bak"
  \rm /tmp/ns-validation-stdout.log /tmp/ns-validation-stderr.log
}

function common_utils() {
  ROOT_DIR="/opt/UiPathAutomationSuite/UiPath_Installer"
  INFILE="${ROOT_DIR}/input.json"
  if [[ ! -s "${INFILE}" ]]; then
    error "Cluster config file not present or empty at $INFILE"
  fi  
  if [[ ! -f "${ROOT_DIR}/Modules/utils.sh" ]]; then
    error "Installer utils file not present at ${ROOT_DIR}/Modules/utils.sh"
  fi
  #shellcheck disable=SC1090,SC1091
  source "${ROOT_DIR}/Modules/utils.sh"
}


function parse_prereq_input_file() {
  CUSTOM_DNS_RESOLVER="$(get_optional_json_value ".infra.custom_dns_resolver")"
}

function execute_name_server_test() {
  local testId="$1"
  local expectedResult="$2"
  local expectedString="$3"
  local result="false"
  local exitCode
  local expectedExitCode

  case "$expectedResult" in
    PASSED|WARN|INFO)
      expectedExitCode=0
      ;;
    FAILED)
      expectedExitCode=1
      ;;
    *)
      error "Unsupported expectedResult=$expectedResult PASSED to ${FUNCNAME[0]}"
      ;;
  esac

  "${ROOT_DIR}/validateUiPathASReadiness.sh" -i "${ROOT_DIR}/input.json" --install-type "${DEPLOYMENT_TYPE}" >/tmp/ns-validation-stdout.log 2>/tmp/ns-validation-stderr.log
  exitCode="$?"
  if grep -q "${expectedResult}.*${expectedString}" /tmp/ns-validation-stdout.log /tmp/ns-validation-stderr.log  && [[ $exitCode -eq $expectedExitCode ]]; then
    result="true"
  fi
  appendOutput "${testId}" "${result}"
  restore_resolver_file
}

function backup_resolverFile() {
  \cp "${DNS_RESOLVER_FILE}" "${DNS_RESOLVER_FILE}.bak"
}

function restore_resolver_file() {
  \cp "${DNS_RESOLVER_FILE}.bak" "${DNS_RESOLVER_FILE}"
}

function execute_tests() {
  ## Run empty / missing resolver config test
  echo -n "" > "${DNS_RESOLVER_FILE}"
  execute_name_server_test "1664078" "FAILED" "nameserver validation failed, ${DNS_RESOLVER_FILE} missing or empty"

  ## Add search domain that responds to * queries
  ## Example: irtt.com
  echo "search irtt.com" >> "${DNS_RESOLVER_FILE}"
  execute_name_server_test "1664079" "FAILED" "Unexpected DNS resolution occurred"

  ## empty nameserver config
  sed -i '/nameserver/d' "${DNS_RESOLVER_FILE}"
  echo "nameserver" >> "${DNS_RESOLVER_FILE}"
  execute_name_server_test "1664080" "FAILED" "nameserver validation failed, nameserver not configured correctly"

  ## Resolver config contains only loopback address
  sed -i '/nameserver/d' "${DNS_RESOLVER_FILE}"
  echo "nameserver 127.0.0.53" >> "${DNS_RESOLVER_FILE}"
  execute_name_server_test "1664081" "FAILED" "Loopback ip address"

  ## Resolver confif contains invalid IP
  sed -i '/nameserver/d' "${DNS_RESOLVER_FILE}"
  echo "nameserver 330.0.0.1" >> "${DNS_RESOLVER_FILE}"
  execute_name_server_test "1664082" "FAILED" "Invalid IP(s) Found"

  ## At-least one nameserver does not resolve FQDN
  echo "nameserver 1.1.1.1" >> "${DNS_RESOLVER_FILE}"
  execute_name_server_test "1664083" "FAILED" "did not resolve via nameserver"

  ## Unreachable nameserver
  echo "nameserver 2.2.2.2" >> "${DNS_RESOLVER_FILE}"
  execute_name_server_test "1664084" "FAILED" "Connection timed out while trying to reach nameserver"

  ## FQDN or Sub-domain not resolvable
  sed -i '/nameserver/d' "${DNS_RESOLVER_FILE}"
  echo "nameserver 1.1.1.1" >> "${DNS_RESOLVER_FILE}"
  execute_name_server_test "1664090" "FAILED" "No IP Address found for"

  ## FQDN and subdomain do not resolvable to same ips
  FQDN=$(jq -r '.fqdn' "${ROOT_DIR}/input.json")
  sed -i "/${FQDN}/d" /etc/hosts
  echo "10.10.10.10 alm.${FQDN}" >> /etc/hosts
  execute_name_server_test "1664089" "FAILED" "the same IP(s) as for the domain"
  # revert changes
  sed -i '/10.10.10.10/d' /etc/hosts
}

function main() {
  common_utils
  parse_prereq_input_file
  DNS_RESOLVER_FILE="/etc/resolv.conf"
  [[ -n "${CUSTOM_DNS_RESOLVER}" ]] && DNS_RESOLVER_FILE=${CUSTOM_DNS_RESOLVER}
  backup_resolverFile
  execute_tests
}

parse_args "$@"
main