#!/bin/bash

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)

if [[ -z "${SCRIPT_DIR}" ]]; then
    echo "Script dir not present"
    exit 1
fi

#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}/utils.sh"

declare -A sub_domains
sub_domains["objectstore"]="all"
sub_domains["monitoring"]="all"
sub_domains["alm"]="all"
sub_domains["insights"]="insights"
sub_domains["registry"]="offline"

INSTALLER_DIR="/opt/UiPathAutomationSuite/UiPath_Installer"

function checkMessageIsAvailable() {
  local logFile="$1"
  local message="$2"
  local messageFindResult result
  result=1
  messageFindResult="$(grep "${message}" ${logFile})"
  if [[ -n "${messageFindResult}" ]]; then
    result=0
  fi
  echo ${result}
}

function update_result() {
  local id="$1"
  local result="$2"
  local rexpectedResult="$3"
  if [[ ${result} == ${rexpectedResult} ]]; then
    appendOutput "${id}" "true" 
  else
    appendOutput "${id}" "false" 
  fi
}

function validationVirtualService() {
  local fqdn="$1"
  local result
  result=0

  allVirtualHosts=$(kubectl get vs -A  --no-headers -o custom-columns=":spec.hosts")
  for virtualHosts in ${allVirtualHosts[@]}; do
    virtualHosts="$(echo ${virtualHosts} | sed -e 's/\[//g' -e 's/\]//g')"
    while IFS=';' read -ra virtualHost; do
      if [[ "${virtualHost}" != *".${fqdn}" && "${virtualHost}" != "${fqdn}" ]]; then
        result=1
        break;
      fi
    done <<< "$virtualHosts"
  done
  echo ${result}
}

function checkIfSubdomainAccessible() {
  local fqdn="$1"
  local result fqdnOutput almOutput monitoringOutput
  result=0
  fqdnOutput=$(curl -m 10  -w "%{http_code}" --silent --output /dev/null "https://${fqdn}")
  almOutput=$(curl -m 10  -w "%{http_code}" --silent --output /dev/null "https://alm.${fqdn}")
  monitoringOutput=$(curl -m 10  -w "%{http_code}" --silent --output /dev/null "https://monitoring.${fqdn}")
  if [[ $fqdnOutput == "000" || $almOutput == "000" || $monitoringOutput == "000" ]] ; then
    result=1
  fi
  echo ${result}
}

function checkIfCertificateAddedVMTrustStore() {
  local fqdn="$1"
  local result certStatus
  result=0
  curl -v https://"${fqdn}" &> /tmp/fqdn_ping.log
  certStatus=$(grep "SSL certificate verify ok." /tmp/fqdn_ping.log)
  if [[ -z ${certStatus} ]]; then
    result=1
  fi
  echo ${result}
}

function updateFQDNShouldGenerateSelfSignCertificateIfNotProvided() {
  local installerRootDir="$1"
  local clusterConfigFile="$2"
  local oldFQDN="$3"
  local newFQDN="$4"
  local prerequisitesArgs=("${@:5}")
  local result newClusterConfigFile virtualServiceStatus certStatus subdomainStatus certFolder

  result=0
  newClusterConfigFile="${installerRootDir}"/update_fqdn_input.json
  certFolder="/tmp/new_cert"
  rm -rf ${certFolder} &> /dev/null
  mkdir -p ${certFolder}

  # create new input file with
  jq ".fqdn = \"${newFQDN}\" | .fixed_rke_address = \"${newFQDN}\" | .server_certificate += {\"ca_cert_file\": \"${certFolder}/rootCA.crt\", \"tls_cert_file\": \"${certFolder}/server.crt\", \"tls_key_file\": \"${certFolder}/server.key\"} " ${clusterConfigFile} > ${newClusterConfigFile}

  # check if cluster is in correct state
  virtualServiceStatus=$(validationVirtualService ${oldFQDN})
  certStatus=$(checkIfCertificateAddedVMTrustStore ${oldFQDN})
  subdomainStatus=$(checkIfSubdomainAccessible ${oldFQDN})

  if [[ "${virtualServiceStatus}" == "1" || "${certStatus}" == "1" || "${subdomainStatus}" == "1" ]]; then
    # skip test run if cluster is not in proper state
    return
  fi

  "${installerRootDir}/install-uipath.sh" -i "${newClusterConfigFile}" ${prerequisitesArgs[@]} &>/tmp/updatefqdn.log
  # Check cluster status post update
  virtualServiceStatus=$(validationVirtualService ${newFQDN})
  certStatus=$(checkIfCertificateAddedVMTrustStore ${newFQDN})
  subdomainStatus=$(checkIfSubdomainAccessible ${newFQDN})

  # update test rail result
  update_result "1619832" "${virtualServiceStatus}" "0"
  update_result "1619826" "${certStatus}" "0"
  update_result "1619830" "${subdomainStatus}" "0"
  
  
  if [[ "${virtualServiceStatus}" == "1" || "${certStatus}" == "1" || "${subdomainStatus}" == "1" ]]; then
    appendOutput "1619828" "false"
  else
    appendOutput "1619828" "true"
  fi
}

function updateFQDNShouldUseUserProvidedCertificate() {
  local installerRootDir="$1"
  local clusterConfigFile="$2"
  local oldFQDN="$3"
  local newFQDN="$4"
  local prerequisitesArgs=("${@:5}")
  local result newClusterConfigFile virtualServiceStatus certStatus subdomainStatus certFolder

  # check if cluster is in correct state
  virtualServiceStatus=$(validationVirtualService ${oldFQDN})
  certStatus=$(checkIfCertificateAddedVMTrustStore ${oldFQDN})
  subdomainStatus=$(checkIfSubdomainAccessible ${oldFQDN})

  if [[ "${virtualServiceStatus}" == "1" || "${certStatus}" == "1" || "${subdomainStatus}" == "1" ]]; then
    # skip test run if cluster is not in proper state
    return
  fi

  result=0
  certFolder="/tmp/new_cert"
  newClusterConfigFile="${installerRootDir}"/update_fqdn_input.json
  rm -rf "${certFolder}" &> /dev/null
  mkdir -p "${certFolder}"
  create_certs "${certFolder}" "${newFQDN},alm.${newFQDN},monitoring.${newFQDN},objectstore.${newFQDN},registry.${newFQDN},insights.${newFQDN}"

  # create new input file with
  jq ".fqdn = \"${newFQDN}\" | .fixed_rke_address = \"${newFQDN}\" | .server_certificate += {\"ca_cert_file\": \"${certFolder}/rootCA.crt\", \"tls_cert_file\": \"${certFolder}/server.crt\", \"tls_key_file\": \"${certFolder}/server.key\"} " ${clusterConfigFile} > ${newClusterConfigFile}

  "${installerRootDir}/install-uipath.sh" -i "${newClusterConfigFile}" ${prerequisitesArgs[@]} &>/tmp/updatefqdn.log
  # Check cluster status post update
  virtualServiceStatus=$(validationVirtualService ${newFQDN})
  certStatus=$(checkIfCertificateAddedVMTrustStore ${newFQDN})
  subdomainStatus=$(checkIfSubdomainAccessible ${newFQDN})
  
  if [[ "${virtualServiceStatus}" == "1" || "${certStatus}" == "1" || "${subdomainStatus}" == "1" ]]; then
    appendOutput "1664188" "false"
  else
    appendOutput "1664188" "true"
  fi
}

function whenFQDNIsResolvableButCertificateIsInvalid_updateShouldStopAtPrereqStage() {
  local installerRootDir="$1"
  local clusterConfigFile="$2"
  local fqdn="$3"
  local deploymentType="$4"
  local is_insights_enabled="$5"
  local prerequisitesArgs=("${@:6}")
  local result newClusterConfigFile certFolder

  certFolder="/tmp/new_cert"
  rm -rf "${certFolder}" &> /dev/null
  mkdir -p "${certFolder}"
  create_certs "/tmp/new_cert" "tmp_${fqdn},*.tmp_${fqdn}"
  result=0
  newClusterConfigFile="${installerRootDir}"/update_fqdn_input.json
  jq ".fqdn = \"${fqdn}\" | .fixed_rke_address = \"${fqdn}\"| .server_certificate += {\"ca_cert_file\": \"${certFolder}/rootCA.crt\", \"tls_cert_file\": \"${certFolder}/server.crt\", \"tls_key_file\": \"${certFolder}/server.key\"} " ${clusterConfigFile} > ${newClusterConfigFile}
  
  "${installerRootDir}/install-uipath.sh" -i "${newClusterConfigFile}" ${prerequisitesArgs[@]} &>/tmp/updatefqdn.log
  for sub_domain in ${sub_domains[@]}; do
    if [[ "${sub_domains[${sub_domain}]}" == "all" || "${sub_domains[${sub_domain}]}" == "${deploymentType}" ]]; then
      result=$(checkMessageIsAvailable  "/tmp/updatefqdn.log" "doesn't have ${sub_domain}.${fqdn} in SAN" )
    elif [[ "${is_insights_enabled}" == "true" && "${sub_domains[${sub_domain}]}" == "insights"  ]]; then
      result=$(checkMessageIsAvailable  "/tmp/updatefqdn.log" "doesn't have insights.${fqdn} in SAN" )
    fi
    if [[ ${result} == 1 ]]; then
      break;
    fi
  done

  update_result "1664186" "${result}" "0"
}

function whenFQDNNotResolvable_updateShouldStopAtPrereqStage() {
  local installerRootDir="$1"
  local clusterConfigFile="$2"
  local fqdn="$3"
  local deploymentType="$4"
  local is_insights_enabled="$5"
  local prerequisitesArgs=("${@:6}")
  local result newClusterConfigFile certFolder

  result=0
  certFolder="/tmp/new_cert"
  rm -rf "${certFolder}" &> /dev/null
  mkdir -p "${certFolder}"
  create_certs "${certFolder}" "${fqdn},*.${fqdn}"
  
  newClusterConfigFile="${installerRootDir}"/update_fqdn_input.json
  jq ".fqdn = \"${fqdn}\" | .fixed_rke_address = \"${fqdn}\"| .server_certificate += {\"ca_cert_file\": \"${certFolder}/rootCA.crt\", \"tls_cert_file\": \"${certFolder}/server.crt\", \"tls_key_file\": \"${certFolder}/server.key\"} " ${clusterConfigFile} > ${newClusterConfigFile}
  
  "${installerRootDir}/install-uipath.sh" -i "${newClusterConfigFile}" ${prerequisitesArgs[@]} &>/tmp/updatefqdn.log
  for sub_domain in ${sub_domains[@]}; do
    if [[ "${sub_domains[${sub_domain}]}" == "all" || "${sub_domains[${sub_domain}]}" == "${deploymentType}" ]]; then
      result=$(checkMessageIsAvailable  "/tmp/updatefqdn.log" "No IP Address found for ${sub_domain}.${fqdn}" )
    elif [[ "${is_insights_enabled}" == "true" && "${sub_domains[${sub_domain}]}" == "insights"  ]]; then
      result=$(checkMessageIsAvailable  "/tmp/updatefqdn.log" "No IP Address found for ${sub_domain}.${fqdn}" )
    fi
    if [[ ${result} == 1 ]]; then
      break;
    fi
  done
  
  update_result "1619823" "${result}" "0"
}

function executeTest() {
  local is_insights_enabled is_offline fqdn
  local prerequisites_args=()
  prerequisites_args+=(-o output.json)
  prerequisites_args+=(--force-update-fqdn)
  prerequisites_args+=(--accept-license-agreement)
  prerequisites_args+=(--skip-compare-config)
  if [[ -z ${NODE_TYPE} ]]; then
    prerequisites_args+=(-a)
  else
    prerequisites_args+=(-k)
    prerequisites_args+=(-k)
  fi

  is_offline="false"
  if [[ "${DEPLOYMENT_TYPE}" == "offline" ]]; then 
    is_offline="true"
  fi

  is_insights_enabled="$(jq ".insights.enabled" "${INSTALLER_DIR}/input.json" | xargs)"
  fqdn="$(jq ".fqdn" "${INSTALLER_DIR}/input.json" | xargs)"
  whenFQDNNotResolvable_updateShouldStopAtPrereqStage "${INSTALLER_DIR}" "${INSTALLER_DIR}/input.json" "random-fqdn.unrechabletest.com" "${is_offline}" "${is_insights_enabled}" "${prerequisites_args[@]}"
  whenFQDNIsResolvableButCertificateIsInvalid_updateShouldStopAtPrereqStage "${INSTALLER_DIR}" "${INSTALLER_DIR}/input.json" "random-fqdn.unrechabletest.com" "${is_offline}" "${is_insights_enabled}" "${prerequisites_args[@]}"
  # make sure this test run at the end becuase and in this sequence only
  updateFQDNShouldGenerateSelfSignCertificateIfNotProvided "${INSTALLER_DIR}" "${INSTALLER_DIR}/input.json" "${fqdn}" "test-dns-${fqdn}" "${prerequisites_args[@]}"
  updateFQDNShouldUseUserProvidedCertificate "${INSTALLER_DIR}" "${INSTALLER_DIR}/input.json" "test-dns-${fqdn}" "${fqdn}" "${prerequisites_args[@]}"
}

parse_args "$@"
executeTest