#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS:
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/upgrade_utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/argocd_utils.sh"

####
# Sets current rook-version and ceph-version from all pods
####
#shellcheck disable=SC2034
function set_rook_ceph_version() {
    local rook_cluster_namespace=rook-ceph
    current_rook_version=$(kubectl -n $rook_cluster_namespace get deployment -l rook_cluster=$rook_cluster_namespace -o jsonpath='{range .items[*]}{"rook-version="}{.metadata.labels.rook-version}{"\n"}{end}' | sort | uniq)
    # current_ceph_version=$(kubectl -n $rook_cluster_namespace get deployment -l rook_cluster=$rook_cluster_namespace -o jsonpath='{range .items[*]}{"ceph-version="}{.metadata.labels.ceph-version}{"\n"}{end}' | sort | uniq)
}

OSD_ALLOWED_TO_CLEANUP=$(kubectl -n rook-ceph get deploy -l app=rook-ceph-osd  -o json | jq -c '[.items[] | .metadata.name]')
if [[ -z "${OSD_ALLOWED_TO_CLEANUP}" || $OSD_ALLOWED_TO_CLEANUP == "[]" ]]; then
    warn "Failure getting OSD deployment. Setting it to default"
    OSD_ALLOWED_TO_CLEANUP='["rook-ceph-osd-0","rook-ceph-osd-1","rook-ceph-osd-2"]'
fi

function cleanup_crashing_osd() {
    local restart_operator="false"
    local min_required_healthy_osd=1
    local in_osd
    local up_osd
    local healthy_osd_pod_count
    local crashed_osd_deploy
    local crashed_pvc_name

    if ! kubectl -n rook-ceph exec deploy/rook-ceph-tools -- ceph osd pool ls detail  | grep 'rook-ceph.rgw.buckets.data' | grep -q 'replicated'; then
        min_required_healthy_osd=2
    fi
    in_osd=$(kubectl -n rook-ceph exec deploy/rook-ceph-tools -- ceph status   -f json  | jq -r '.osdmap.num_in_osds')
    up_osd=$(kubectl -n rook-ceph exec deploy/rook-ceph-tools -- ceph status   -f json  | jq -r '.osdmap.num_up_osds')
    healthy_osd_pod_count=$(kubectl -n rook-ceph get pod -l app=rook-ceph-osd | grep 'Running' | grep -c -P '([0-9])/\1')
    if ! [[ $in_osd -ge $min_required_healthy_osd && $up_osd -ge $min_required_healthy_osd && $healthy_osd_pod_count -ge $min_required_healthy_osd ]]; then
        return
    fi
    for crashed_osd_deploy in $(kubectl -n rook-ceph get pod -l app=rook-ceph-osd  | grep 'CrashLoopBackOff' | cut -d'-' -f'1-4') ; do
        if kubectl -n rook-ceph logs "deployment/${crashed_osd_deploy}" | grep -q '/crash/'; then
            echo "Found crashing OSD deployment: '${crashed_osd_deploy}'"
            if ! jq -n  -c --argjson allowedOSDs "${OSD_ALLOWED_TO_CLEANUP}"  --arg osdDeploy "${crashed_osd_deploy}" '$allowedOSDs | index($osdDeploy)' | grep -q -E '^[0-9]+$'; then
                error "Won't try recovery of OSD deployment: '${crashed_osd_deploy}' as it is not part of allowed list: '${OSD_ALLOWED_TO_CLEANUP}'. Please work with Uipath support to complete the upgrade"
            fi
            crashed_pvc_name=$(kubectl -n rook-ceph get deployment "${crashed_osd_deploy}" -o json | jq -r '.metadata.labels["ceph.rook.io/pvc"]')
            info "Removing crashing OSD deployment: '${crashed_osd_deploy}' and PVC: '${crashed_pvc_name}'"
            timeout 60  kubectl -n rook-ceph delete deployment "${crashed_osd_deploy}" || kubectl -n rook-ceph delete deployment "${crashed_osd_deploy}" --force --grace-period=0
            timeout 100 kubectl -n rook-ceph delete pvc "${crashed_pvc_name}" || kubectl -n rook-ceph delete pvc "${crashed_pvc_name}" --force --grace-period=0
            OSD_ALLOWED_TO_CLEANUP=$(jq -n  -c --argjson allowedOSDs "${OSD_ALLOWED_TO_CLEANUP}" --arg osdDeploy "${crashed_osd_deploy}"  '$allowedOSDs | del(.[] | select(. == $osdDeploy))')
            restart_operator="true"
        fi
    done
    if [[ $restart_operator == "true" ]]; then
        kubectl -n rook-ceph rollout restart deployment/rook-ceph-operator
    fi
    return 0
}

function validate_rook_ceph_version() {
    local return_code=0
    local rook_cluster_namespace=rook-ceph
    local required_rook_version="sf-rook-ceph-fork:v1.8.9-4"
    local required_ceph_version="ceph/ceph:v15.2.10"

    kubectl -n "${rook_cluster_namespace}" get deploy rook-ceph-operator -o json | jq -r '.spec.template.spec.containers[0].image' | grep -q -v -F "${required_rook_version}" && return_code=1
    kubectl -n "${rook_cluster_namespace}" get deploy -l rook_cluster="${rook_cluster_namespace}" -o json | jq -r '.items[].spec.template.spec.containers[0].image' | grep -q -v -F "${required_ceph_version}" && return_code=1
    return "${return_code}"
}

####
# Waits for all rook-ceph pods to be updated to latest version and checks ceph health in the end
####
function check_latest_ceph_health() {
    
    if validate_rook_ceph_version;
    then
        # If all pods are successfully updated, check ceph health
        info "Rook ceph already upgraded to latest version"
        return 0
    fi

    osd_deploy_orig=$(kubectl -n rook-ceph get cephcluster rook-ceph -o json | jq '.spec.mon.count')
    
    local rook_cluster_namespace=rook-ceph
    local try=0
    local maxtry=100
    info "Waiting for rook-ceph to be updated to latest version.. ${try}/${maxtry}"
    while ((try != maxtry));
    do
        osd_count=$(kubectl -n rook-ceph get deploy -l app=rook-ceph-osd --no-headers 2>/dev/null | wc -l)
        if [[ "${osd_count}" -ge "${osd_deploy_orig}" ]];
        then
            if validate_rook_ceph_version;
            then
                info "Rook-Ceph updated to latest version"
                break
            fi
        fi
        try=$((try+1))
        cleanup_crashing_osd
        info "Waiting for rook-ceph to be updated to latest version.. ${try}/${maxtry}"
        sleep 30
    done

    osd_deployments=$(kubectl -n rook-ceph get deploy -l app=rook-ceph-osd -o name)
    for deploy in ${osd_deployments};
    do
        kubectl -n rook-ceph rollout status "$deploy" --timeout=3m || cleanup_crashing_osd
    done

    kubectl -n rook-ceph delete pods -l app=rook-ceph-rgw
    gateway_deploy=$(kubectl -n rook-ceph get deploy -l app=rook-ceph-rgw -o name)
    for deploy in ${gateway_deploy};
    do
        kubectl -n rook-ceph rollout status "$deploy" --timeout=3m || true
    done

    if validate_rook_ceph_version;
    then
        # If all pods are successfully updated, check ceph health
        ceph_health_exit_on_error
        info "Rook ceph upgrade successful"
    else
        # As all pods could not be updated to latest version, exit
        kubectl -n $rook_cluster_namespace get deployments -l rook_cluster=$rook_cluster_namespace -o jsonpath='{range .items[*]}{.metadata.name}{"  \treq/upd/avl: "}{.spec.replicas}{"/"}{.status.updatedReplicas}{"/"}{.status.readyReplicas}{"  \trook-version="}{.metadata.labels.rook-version}{"\n"}{end}'
        error "Rook-Ceph could not be updated to latest version"
    fi
}

check_latest_ceph_health
